
# Bond 
china_bondflow_fields = ["日期", "余额", "利息", "本金", "执行利率", "本息合计", "备注"]
china_bondflow_fields_s = ["余额", "利息", "本金", "执行利率", "本息合计", "备注"]
china_bond_cashflow = ["本金","利息","本息合计"]
english_bondflow_fields = ["date", "balance", "interest", "principal", "rate", "cash", "memo"]
english_bondflow_fields_s = ["balance", "interest", "principal", "rate", "cash", "memo"]

# Pool
## mortgage
china_mortgage_flow_fields = ["余额", "本金", "利息", "早偿金额", "违约金额", "回收金额", "损失金额", "利率", "债务人数量"]
english_mortgage_flow_fields = ["Balance", "Principal", "Interest", "Prepayment", "Default", "Recovery", "Loss", "WAC","BorrowerNum"]
china_mortgage_flow_fields_d = ["日期"]+ china_mortgage_flow_fields
english_mortgage_flow_fields_d = ["Date"] + english_mortgage_flow_fields
## Rental
china_rental_flow = ['待收金额','租金']
english_rental_flow = ['Balance','Rental']
china_rental_flow_d = ["日期"] + china_rental_flow
english_rental_flow_d = ["Date"] + english_rental_flow
## Loan flow
china_loan_flow = ["余额", "本金", "利息", "早偿金额", "违约金额", "回收金额", "损失金额", "利率"]
english_loan_flow = ["Balance", "Principal", "Interest", "Prepayment", "Default", "Recovery", "Loss", "WAC"]
china_loan_flow_d = ["日期"] + china_loan_flow
english_loan_flow_d = ["Date"] + english_loan_flow

# Fee 
china_fee_flow_fields_d = ["日期", "余额", "支付", "剩余支付", "备注"]
english_fee_flow_fields_d = ["date", "balance", "payment", "due", "memo"]

# Account
china_acc_flow_fields_d = ["日期", "余额", "变动额", "备注"]
english_acc_flow_fields_d = ["date", "balance", "change", "memo"]


# LiqProvider
china_liq_flow_fields_d = ["日期", "限额", "变动额", "已提供","利息","费用","备注"]
english_liq_flow_fields_d = ["date", "balance", "change", "used","int","premium","memo"]

# Rate Swap 
china_rs_flow_fields_d = ["日期", "面额", "变动额", "支付","收取","净额","备注"]
english_rs_flow_fields_d = ["date", "balance", "amount", "pay","receive","due","memo"]

# Ledger
china_ledger_flow_fields_d = ["日期", "余额", "变动额", "备注"]
english_ledger_flow_fields_d = ["date", "balance", "amount", "comment"]

# Index



# deal status



datePattern = {"月末": "MonthEnd", "季度末": "QuarterEnd", "年末": "YearEnd", "月初": "MonthFirst",
               "季度初": "QuarterFirst", "年初": "YearFirst", "每年": "MonthDayOfYear", "每月": "DayOfMonth", "每周": "DayOfWeek"}


freqMap = {"每月": "Monthly", "每季度": "Quarterly", "每半年": "SemiAnnually", "每年": "Annually", "Monthly": "Monthly", "Quarterly": "Quarterly", "SemiAnnually": "SemiAnnually", "Annually": "Annually", "monthly": "Monthly", "quarterly": "Quarterly", "semiAnnually": "SemiAnnually", "annually": "Annually"
           }

baseMap = {"资产池余额": "CurrentPoolBalance"
           , "资产池期末余额": "CurrentPoolBalance"
           , "资产池期初余额": "CurrentPoolBegBalance"
           , "资产池初始余额": "OriginalPoolBalance"
           , "初始资产池余额": "OriginalPoolBalance"
           , "资产池当期利息": "PoolCollectionInt"
           , "债券余额": "CurrentBondBalance"
           , "债券初始余额": "OriginalBondBalance"
           , "当期已付债券利息": "LastBondIntPaid"
           , "当期已付费用": "LastFeePaid"
           , "当期未付债券利息": "CurrentDueBondInt"
           , "当期未付费用": "CurrentDueFee"}
           
           
#pool income mapping
poolSourceMapping = {"利息回款": "CollectedInterest"
                    , "本金回款": "CollectedPrincipal"
                    , "早偿回款": "CollectedPrepayment"
                    , "回收回款": "CollectedRecoveries"
                    , "租金回款": "CollectedRental"
                    , "CollectedInterest": "CollectedInterest"
                    , "CollectedPrincipal": "CollectedPrincipal"
                    , "CollectedPrepayment": "CollectedPrepayment"
                    , "CollectedRecoveries": "CollectedRecoveries"
                    , "CollectedRental": "CollectedRental"
                     }

op_map = {
    ">":"G"
    ,">=":"GE"
    ,"<":"L"
    ,"<=":"LE"
    ,"=":"E"
}

dealStatusMap = {"摊还": "Current", "加速清偿": "Accelerated", "循环": "Revolving"}


#Deal Cycle
chinaDealCycle = {"回收后":"EndCollection"
                 ,"回收动作后":"EndCollectionWF"
                 ,"分配前":"BeginDistributionWF"
                 ,"分配后":"EndDistributionWF"
                 ,"分配中":"InWF"}

englishDealCycle = {"BeforeCollect":"EndCollection"
                 ,"AfterCollect":"EndCollectionWF"
                 ,"BeforeDistribution":"BeginDistributionWF"
                 ,"AfterDistribution":"EndDistributionWF"
                 ,"InDistribution":"InWF"}

#Asset pricing 
assetPricingHeader = {
    "chinese":["估值","WAL","Duration","Convexity","AccruedInterest"]
    ,"english":["Pricing","WAL","Duration","Convexity","AccruedInterest"]
}


import pandas as pd


def comp_df(x, y):
    return pd.merge(x, y, on="日期", how='outer').sort_index().sort_index(axis=1)

def comp_engines(engine1,engine2, d, a=None):
    
    r1 = engine1.run(d,assumptions=a,read=True)
    r2 = engine2.run(d,assumptions=a,read=True)

    comp_result = {}
    # pool check
    if not r1['pool']['flow'].equals(r2['pool']['flow']):
        comp_result['pool'] = comp_df(r1['pool']['flow'],r2['pool']['flow'])
    else:
        comp_result['pool'] = True

    # expense check  
    comp_result['fee'] = {}
    for fn,f in r1['fees'].items():
        if not f.equals(r2['fees'][fn]):
            comp_result['fee'][fn] = comp_df(f,r2['fees'][fn])
        else:
            comp_result['fee'][fn] = True

    # bond check
    comp_result['bond'] = {}
    for fn,f in r1['bonds'].items():
        if not f.equals(r2['bonds'][fn]):
            comp_result['bond'][fn] = comp_df(f,r2['bonds'][fn])
        else:
            comp_result['bond'][fn] = True


    # account check
    comp_result['account'] = {}
    for fn,f in r1['accounts'].items():
        if not f.equals(r2['accounts'][fn]):
            comp_result['account'][fn] = comp_df(f,r2['accounts'][fn])
        else:
            comp_result['account'][fn] = True

    return comp_result



import json, datetime, pickle, re, urllib3, getpass, copy
from importlib.metadata import version
from json.decoder import JSONDecodeError
from dataclasses import dataclass,field

import rich
from rich.console import Console
from rich.json import JSON
import requests
from requests.exceptions import ConnectionError,ReadTimeout
import pandas as pd
from pyspecter import query

from absbox.local.util import mkTag, isDate, flat, guess_pool_locale, mapValsBy, guess_pool_flow_header, _read_cf, _read_asset_pricing, mergeStrWithDict
from absbox.local.component import mkPool, mkAssumption, mkAssumption2, mkPricingAssump,mkLiqMethod,mkAssetUnion
from absbox.local.base import *
from absbox.validation import valReq,valAssumption

from absbox.local.china import SPV
from absbox.local.generic import Generic


VERSION_NUM = version("absbox")
urllib3.disable_warnings()
console = Console()

@dataclass
class API:
    url: str
    lang: str = "chinese"
    check: bool = True
    server_info = {}
    version = VERSION_NUM.split(".")
    hdrs = {'Content-type': 'application/json', 'Accept': 'text/plain','Accept':'*/*' ,'Accept-Encoding':'gzip'}
    session = None

    def __post_init__(self):
        self.url = self.url.rstrip("/")
        with console.status(f"[magenta]Connecting engine server -> {self.url}") as status:
            try:
                _r = requests.get(f"{self.url}/version",verify=False, timeout=5).text
            except (ConnectionRefusedError, ConnectionError):
                console.print(f"❌[bold red]Error: Can't not connect to API server {self.url}")
                self.url = None
                return

            self.server_info = self.server_info | json.loads(_r)
            engine_version = self.server_info['_version'].split(".")
            if self.check and (self.version[1] != engine_version[1]):
                console.print(f"❌[bold red]Failed to init the api instance, lib support={self.version} but server version={self.server_info['_version']} , pls upgrade your api package by: pip -U absbox")
                return
        console.print(f"✅[bold green]Connected, local lib:{'.'.join(self.version)}, server:{'.'.join(engine_version)}")
        self.session = requests.Session() 

    def build_req(self, deal, assumptions=None, pricing=None) -> str:
        r = None
        _assump = None 
        _deal = deal.json if hasattr(deal,"json") else deal
        _pricing = mkPricingAssump(pricing) if pricing else None
        if isinstance(assumptions, dict):
            _assump = mapValsBy(assumptions, mkAssumption2)
            r = mkTag(("MultiScenarioRunReq",[_deal, _assump, _pricing]))
        elif isinstance(assumptions, list) :   
            _assump = mkAssumption2(assumptions)
            r = mkTag(("SingleRunReq",[_deal, _assump, _pricing]))
        elif assumptions is None:
            r = mkTag(("SingleRunReq",[_deal, None, _pricing]))
        return json.dumps(r , ensure_ascii=False)

    def build_pool_req(self, pool, assumptions, read=None) -> str:
        r = None
        if isinstance(assumptions, list):
            r = mkTag(("SingleRunPoolReq"
                      ,[mkPool(pool)
                      ,mkAssumption2(assumptions)])) 
        elif isinstance(assumptions, dict):
            r = mkTag(("MultiScenarioRunPoolReq"
                      ,[mkPool(pool)
                      ,mapValsBy(assumptions, mkAssumption2)])) 
        else:
            raise RuntimeError("Error in build pool req")
        return json.dumps(r , ensure_ascii=False)

    def validate(self, _r) -> list:
        # validatin waterfall
        error, warning = valReq(_r)
        if warning:
            console.print(f"❕[bold yellow]Warning in model :{warning}")
        if len(error)>0:
            console.print(f"❌[bold red]Error in model :{error}")
            return False,error,warning
        else:
            return True,error,warning

    def run(self, deal,
            assumptions=None,
            pricing=None,
            read=True):

        # if run req is a multi-scenario run
        multi_run_flag = True if isinstance(assumptions, dict) else False
        url = f"{self.url}/runDealByScenarios"  if multi_run_flag  else f"{self.url}/runDeal"

        # construct request
        req = self.build_req(deal, assumptions, pricing)
        #validate deal
        val_result, err, warn = self.validate(req)
        if not val_result:
            return val_result, err, warn
        if pricing is None:
            result = self._send_req(req,url)
        else:
            result = self._send_req(req,url,timeout=30)
        if result is None:
            console.print("❌[bold red]Failed to get response from run")
            return None
        if read and multi_run_flag:
            return {n:deal.read(_r) for (n,_r) in result.items()}
        elif read :
            return deal.read(result)
        else:
            return result

    def runPool(self, pool, assumptions,read=True):
        def read_single(pool_resp):
            flow_header,idx = guess_pool_flow_header(pool_resp[0],pool_lang)
            try:
                result = pd.DataFrame([_['contents'] for _ in pool_resp] , columns=flow_header)
            except ValueError as e:
                console.print(f"❌[bold red]Failed to match header:{flow_header} with {result[0]['contents']}")
            result = result.set_index(idx)
            result.index.rename(idx, inplace=True)
            result.sort_index(inplace=True)
            return result
            
        multi_scenario = True if isinstance(assumptions, dict) else False 
        url = f"{self.url}/runPoolByScenarios" if multi_scenario else f"{self.url}/runPool"
        pool_lang = guess_pool_locale(pool)
        req = self.build_pool_req(pool, assumptions=assumptions)
        result = self._send_req(req, url)

        if read and (not multi_scenario):
            return read_single(result)
        elif read and multi_scenario : 
            return mapValsBy(result, read_single)
        else:
            return result
    
    def runStructs(self, deals, assumptions=None, pricing=None, read=True ):
        assert isinstance(deals, dict),f"Deals should be a dict but got {deals}"
        url = f"{self.url}/runMultiDeals" 
        _assumptions = mkAssumption2(assumptions) if assumptions else None 
        _pricing = mkPricingAssump(pricing) if pricing else None
        req = json.dumps(mkTag(("MultiDealRunReq"
                                ,[{k:v.json for k,v in deals.items()}
                                  ,_assumptions
                                  ,_pricing]))
                        ,ensure_ascii=False)

        result = self._send_req(req,url)
        if read :
            return {k:deals[k].read(v) for k,v in result.items()}    
        else:
            return result

    def runAsset(self, date, _assets, assumptions=None, pricing=None, read=True):
        assert isinstance(_assets, list),f"Assets passed in must be a list"
        def readResult(x):
            (cfs,pr) = x
            cfs = _read_cf(cfs, self.lang)
            pricingResult = _read_asset_pricing(pr, self.lang) if pr else None
            return (cfs,pricingResult)
        url = f"{self.url}/runAsset"
        _assumptions = mkAssumption2(assumptions) 
        _pricing =  mkLiqMethod(pricing) if pricing else None
        assets = [ mkAssetUnion(_) for _ in _assets] 
        req = json.dumps([date ,assets ,_assumptions ,_pricing]
                         ,ensure_ascii=False)
        result = self._send_req(req,url)
        if read :
            return readResult(result)
        else:
            return result

    def loginLibrary(self, user, pw, **q):
        deal_library_url = q['deal_library']+"/token"
        cred = {"user":user,"password":pw}
        r = self._send_req(json.dumps(cred), deal_library_url)
        if 'token' in r:
            console.print(f"✅[bold green] login successfully,{r['msg']}")
            self.token = r['token']
        else:
            if hasattr(self,'token'):
                delattr(self,'token')
            console.print(f"❌[bold red]Failed to login,{r['msg']}")
            return None
    
    def safeLogin(self, user, **q):
        try:
            pw = getpass.getpass()
            self.loginLibrary(user, pw, **q)
        except Exception as e:
            console.print(f"❌[bold red]{e}")
    
    def queryLibrary(self,ks,**q):
        if not hasattr(self,"token"):
            console.print(f"❌[bold red] No token found , please call loginLibrary() to login")
            return 
        deal_library_url = q['deal_library']+"/query"
        d = {"bond_id": [k for k in ks] }
        q = {"read":True} | q
        result = self._send_req(json.dumps(d|q), deal_library_url,headers= {"Authorization":f"Bearer {self.token}"})
        console.print(f"✅[bold green] query success")
        if q['read'] == True:
            if 'data' in result:
                return pd.DataFrame(result['data'],columns=result['header'])
            elif 'error' in result:
                return pd.DataFrame([result["error"]],columns=["error"])
        else:
            return result

    def listLibrary(self,**q):
        deal_library_url = q['deal_library']+"/list"
        result = self._send_req(json.dumps(q), deal_library_url)
        console.print(f"✅[bold green]list success")
        if ('read' in q) and (q['read'] == True):
            return pd.DataFrame(result['data'],columns=result['header'])
        else:
            return result

    def runLibrary(self,_id,**p):
        deal_library_url = p['deal_library']+"/run"
        read = p.get("read",True)
        pricingAssump = p.get("pricing",None)
        dealAssump = p.get("assump",None)
        prod_flag = {"production":p.get("production",True)}
        runReq = mergeStrWithDict (self.build_req(_id, dealAssump, pricingAssump), prod_flag )
        if not hasattr(self,"token"):
            console.print(f"❌[bold red] No token found , please call loginLibrary() to login")
            return 
        result = self._send_req(runReq, deal_library_url, headers={"Authorization":f"Bearer {self.token}"})
        def lookupReader(x):
            match x:
                case "china.SPV":
                    return SPV
                case "generic.Generic":
                    return Generic
                case _:
                    raise RuntimeError(f"Failed to match reader:{x}")
        try:
            ri = result['run_info']
            result = result['run_result']
            console.print(f"✅[bold green]run success with deal id={ri['deal_id']}/report num={ri['report_num']},doc_id={ri['doc_id']}")
        except Exception as e:
            console.print(f"❌[bold red]message from API server:{result}")
            return None
        try:       
            classReader = lookupReader(p['reader'])
            if read and isinstance(result,list):
                return classReader.read(result)
            elif read and isinstance(result, dict):
                return {k:classReader.read(v) for k,v in result.items()}
            else:
                return result
        except Exception as e:
            console.print(f"❌[bold red]{e}")
            return None

    def _send_req(self,_req,_url,timeout=10,headers={})->dict:
        with console.status("") as status:
            try:
                hdrs = self.hdrs | headers
                r = self.session.post(_url, data=_req.encode('utf-8'), headers=hdrs, verify=False, timeout=timeout)
            except (ConnectionRefusedError, ConnectionError):
                console.print(f"❌[bold red] Failed to talk to server {_url}")
                return None
            except ReadTimeout:
                console.print(f"❌[bold red] Failed to get response from server")
                return None
            if r.status_code != 200:
                console.print_json(_req)
                print(">>>",r.text)
                console.print_json(r.text)
                return None
            try:
                return json.loads(r.text)
            except JSONDecodeError as e:
                console.print(e)
                return None
    



import json
from pyspecter import S,query
from absbox.local.util import flat


def inSet(xs,validSet):
    if set(xs).issubset(validSet):
        return True
    else:
        return False

def valDeal(d, error, warning) -> list:
    acc_names = set(d['accounts'].keys())
    bnd_names = set(d['bonds'].keys())
    fee_names = set(d['fees'].keys())
    w = d['waterfall']
    #optional
    if d['ledgers']:
        ledger_names = set(d['ledgers'].keys())
    assert w is not None ,"Waterfall is None"

    for wn,waterfallActions in w.items():
        for idx,(pre,action) in enumerate(waterfallActions):
            match action:
                case {"tag":'PayFeeBy', "contents":[_, accs, fees]}:
                    if (not inSet(accs,acc_names)) or (not inSet(fees,fee_names)):
                        error.append(f"{wn},{idx}")
                case {"tag":'PayFee', "contents":[accs, fees]}:
                    if (not inSet(accs,acc_names)) or (not inSet(fees,fee_names)):
                        error.append(f"{wn},{idx}")     
                case {"tag":'PayInt', "contents":[acc, bonds]}:
                    if (not inSet([acc],acc_names)) or (not inSet(bonds,bnd_names)):
                        error.append(f"{wn},{idx}")  
                case {"tag":'PayPrin', "contents":[acc, bonds]}:
                    if (not inSet([acc],acc_names)) or (not inSet(bonds,bnd_names)):
                        error.append(f"{wn},{idx}")  
                case {"tag":'PayPrinBy', "contents":[_, acc, bond]}:
                    if (not inSet([acc],acc_names)) or (not inSet([bond],bnd_names)):
                        error.append(f"{wn},{idx}")  
                case {"tag":'PayResidual', "contents":[_, acc, bond]}:
                    if (not inSet([acc],acc_names)) or (not inSet([bond],bnd_names)):
                        error.append(f"{wn},{idx}")  
                case {"tag":'Transfer',"contents":[acc1,acc2]}:
                    if (not inSet([acc1,acc2],acc_names)):
                        error.append(f"{wn},{idx}")
                case {"tag":'TransferBy',"contents":[_,acc1,acc2]}:
                    if (not inSet([acc1,acc2],acc_names)):
                        error.append(f"{wn},{idx}")
                case {"tag":'PayTillYield',"contents":[acc, bonds]}:
                    if (not inSet([acc],acc_names)) or (not inSet(bonds,bnd_names)):
                        error.append(f"{wn},{idx}")
                case {"tag":'PayFeeResidual',"contents":[_,acc,fee]}:
                    if (not inSet([acc],acc_names)) or (not inSet([fee],fee_names)):
                        error.append(f"{wn},{idx}")
                case _:
                    pass
    
    return (error,warning)

def valReq(reqSent) -> list:
    error = []
    warning = []
    req = json.loads(reqSent)
    match req :
        case {"tag":"SingleRunReq","contents":[{"contents":d}, ma, mp]}:
            error, warning = valDeal(d, error, warning)
            error, warning = valAssumption(d, ma, error, warning)
        case {"tag":"MultiScenarioRunReq","contents":[{"contents":d}, mam, mp]}:
            error, warning = valDeal(d, error, warning)
        case {"tag":"MultiDealRunReq","contents":[dm, ma, mp]}:
            error, warning = valAssumption(ma, error, warning)
        case _:
            raise RuntimeError(f"Failed to match request:{req}")

    return error, warning

def valAssumption(d, ma , error, warning) -> list:
    def _validate_single_assump(z):
        match z:
            case {'tag': 'PoolLevel'}:
                return [],[]
            case {'tag': 'ByIndex', 'contents':[assumps, _]}:
                _e = []
                _w = []
                _ids = set(flat([ assump[0] for assump in assumps ]))
                if not _ids.issubset(asset_ids):
                    _e.append(f"Not Valid Asset ID:{_ids - asset_ids}")
                if len(missing_asset_id := asset_ids - _ids) > 0:
                    _w.append(f"Missing Asset to set assumption:{missing_asset_id}")            
                return _e,_w
            case _:
                raise RuntimeError(f"Failed to match:{z}")
    
    asset_ids = set(range(len(query(d, ['pool', 'assets']))))
    
    if ma is None:
        return error,warning
    else:
        e,w = _validate_single_assump(ma)
        return error+e,warning+w



from dataclasses import dataclass
import functools

from absbox import *
from absbox.local.util import mkTag,mapListValBy,mapValsBy,renameKs2,guess_pool_flow_header,positionFlow
from absbox.local.component import *
from absbox.local.base import * 
import pandas as pd
import collections


@dataclass
class Generic:
    name: str
    dates: dict
    pool: dict
    accounts: tuple
    bonds: tuple
    fees: tuple
    waterfall: dict
    collection: list
    liqFacility :dict = None
    rateSwap:dict = None
    currencySwap:dict = None
    trigger:dict = None
    status:str = "Amortizing"
    custom: dict = None
    ledgers:dict = None

    @property
    def json(self):
        dists,collects,cleans = [ self.waterfall.get(wn,[]) for wn in ['Normal','PoolCollection','CleanUp']]
        distsAs,collectsAs,cleansAs = [ [ mkWaterfall2(_action) for _action in _actions] for _actions in [dists,collects,cleans] ]
        distsflt,collectsflt,cleanflt = [ itertools.chain.from_iterable(x) for x in [distsAs,collectsAs,cleansAs] ]
        parsedDates = mkDate(self.dates)
        (lastAssetDate,lastCloseDate) = getStartDate(self.dates)
        """
        get the json formatted string
        """
        _r = {
            "dates": parsedDates,
            "name": self.name,
            "status":mkTag((self.status)),
            "pool": {"assets": [mkAsset(x) for x in self.pool.get('assets', [])]
                     , "asOfDate": lastAssetDate
                     , "issuanceStat": self.pool.get("issuanceStat")
                     , "futureCf":mkCf(self.pool.get('cashflow', [])) },
            "bonds": { bn: mkBnd(bn,bo) for (bn,bo) in self.bonds},
            "waterfall": mkWaterfall({},self.waterfall.copy()),  
            "fees": {fn: mkFee(fo|{"name":fn},fsDate = lastCloseDate) 
                                 for (fn,fo) in self.fees},
            "accounts": {an:mkAcc(an,ao) for (an,ao) in self.accounts},
            "collects": [ mkCollection(c) for c in self.collection],
            "rateSwap": { k:mkRateSwap(v) for k,v in self.currencySwap.items()} if self.rateSwap else None,
            "currencySwap":None ,
            "custom": {cn:mkCustom(co) for cn,co in self.custom.items()} if self.custom else None ,
            "triggers": renameKs2(mapListValBy(self.trigger,mkTrigger),englishDealCycle) if self.trigger else None,
            "liqProvider": {ln: mkLiqProvider(ln, lo | {"start":lastCloseDate} ) 
                               for ln,lo in self.liqFacility.items() } if self.liqFacility else None,
            "ledgers": {ln: mkLedger(ln, v) for ln,v in self.ledgers.items()} if self.ledgers else None
        }

        _dealType = identify_deal_type(_r)

        return mkTag((_dealType,_r))

    def read_assump(self, assump):
        if assump:
            return [mkAssumption(a) for a in assump]
        return None

    def read_pricing(self, pricing):
        if pricing:
            return mkPricingAssump(pricing)
        return None
    
    @staticmethod
    def read(resp):
        read_paths = {'bonds': ('bndStmt' , english_bondflow_fields , "bond")
                     , 'fees': ('feeStmt' , english_fee_flow_fields_d , "fee")
                     , 'accounts': ('accStmt' , english_acc_flow_fields_d , "account")
                     , 'liqProvider': ('liqStmt', english_liq_flow_fields_d, "")
                     , 'rateSwap': ('rsStmt', english_rs_flow_fields_d, "")
                     , 'ledgers': ('ledgStmt', english_ledger_flow_fields_d, "")
                     }
        deal_content = resp[0]['contents']
        output = {}
        for comp_name, comp_v in read_paths.items():
            if deal_content[comp_name] is None:
                continue
            output[comp_name] = {}
            for k, x in deal_content[comp_name].items():
                ir = None
                if x[comp_v[0]]:
                    ir = [_['contents'] for _ in x[comp_v[0]]]
                output[comp_name][k] = pd.DataFrame(ir, columns=comp_v[1]).set_index("date")
            output[comp_name] = collections.OrderedDict(sorted(output[comp_name].items()))
        # aggregate fees
        output['fees'] = {f: v.groupby('date').agg({"balance": "min", "payment": "sum", "due": "min"})
                          for f, v in output['fees'].items()}

        # aggregate accounts
        output['agg_accounts'] = aggAccs(output['accounts'], 'en')

        output['pool'] = {}
        if deal_content['pool']['futureCf'] is None:
            output['pool']['flow'] = None
        else:
            _pool_cf_header,_ = guess_pool_flow_header(deal_content['pool']['futureCf'][0],"english")
            output['pool']['flow'] = pd.DataFrame([_['contents'] for _ in deal_content['pool']['futureCf']]
                                                  , columns=_pool_cf_header)
            pool_idx = 'Date'
            output['pool']['flow'] = output['pool']['flow'].set_index(pool_idx)
            output['pool']['flow'].index.rename(pool_idx, inplace=True)

        output['pricing'] = readPricingResult(resp[3], 'en')
        output['result'] = readRunSummary(resp[2], 'en')
        output['_deal'] = resp[0]
        return output



from collections import namedtuple
import pandas as pd 
import numpy as np
from client import API


Portfolio = namedtuple('Portfolio', ['positions'])

#def addPosition(p:Portfolio, bnd, size):
#    pass
#
#
#def calcStats(api:API, p:Portfolio, assumps):
#    ps = p.positions
#    result = {}
#    for p in ps:
#        assump = assumps.get(p)
#        pos = getPos(p)
#        r = api.run(p,assumptions=assump,position=pos)
#        result[p] = r
#    
#    return result
#
#
#def hypoBuy(p:Portfolio, assumps, (bnd, size, pAssump)):
#    currentStats = calcStats(api, p, assumps)
#
#    newAssump = assumps | pAssump
#    newP = p.add((bnd,size))
#    
#    futureStats = calcStats(api, p, assumps)
#    
#
#
#def hypoSell(p:Portfolio, bnd, size):
#    pass


import pandas as pd 



def run_yield_table(api, d, bondName, p_assumps:dict, b_assumps:dict):
    
    rs = api.run(d ,assumptions=p_assumps
                   ,pricing = b_assumps
                   ,read=True)

    b_pricing = {k:v['pricing'].loc[bondName] for k,v in rs.items() }
    b_table = pd.concat(b_pricing.values(),axis=1)
    b_table.columns = list(b_pricing.keys())

    return b_table



import pandas as pd
import functools,json,copy
import logging
import itertools,re
from enum import Enum
import numpy as np
import dataclasses
from functools import reduce
from pyxirr import xirr,xnpv

from absbox.local.base import *

def flat(xss) -> list:
    return reduce(lambda xs, ys: xs + ys, xss)

def mkTag(x) -> dict:
    match x:
        case (tagName, tagValue):
            return {"tag": tagName, "contents": tagValue}
        case (tagName):
            return {"tag": tagName}

def readTagStr(x:str):
    _x = json.loads(x.replace("'","\""))
    match _x:
        case {"tag":_t,"contents":_c} if isinstance(_c, list):
            _cs = [str(_) for _ in _c]
            return f"<{_t}:{','.join(_cs)}>"
        case {"tag": _t }:
            return f"<{_t}>"
        case _ :
            return f"<{_x}>"


def readTag(x:dict):
    return f"<{x['tag']}:{','.join(x['contents'])}>"

def isDate(x):
    return re.match(r"\d{4}\-\d{2}\-\d{2}",x)

def allList(xs):
    return all([ isinstance(x,list)  for x in xs])

def mkTs(n, vs):
    return mkTag((n, vs))

def unify(xs, ns):
    "union dataframes by stacking up with names provided"
    index_name = xs[0].index.name
    dfs = []
    for x, n in zip(xs, ns):
        dfs.append(pd.concat([x], keys=[n], axis=1))
    r = functools.reduce(lambda acc, x: acc.merge(x
                                                , how='outer'
                                                , on=[index_name])
                        , dfs)
    return r.sort_index()

def unifyTs(xs):
    "unify time-series alike dataframe"
    _index_set = set([x.index.name for x in xs])
    assert len(_index_set)==1,f"Index of dataframes should have same name,got:{_index_set}"

    _index_name = list(_index_set)[0]
    r = functools.reduce(lambda acc, x: acc.merge(x
                                                , how='outer'
                                                , on=[_index_name]), xs)
    
    return r.sort_index()

def consolStmtByDate(s):
    return s.groupby("日期").last()

def aggStmtByDate(s):
    return s.groupby("日期").sum()

def aggCFby(_df, interval, cols):
    df = _df.copy()
    idx = None
    dummy_col = '_index'
    df[dummy_col] = df.index
    _mapping = {"月份":"M"
               ,"Month":"M"
               ,"M":"M"
               ,"month":"M"}
    if df.index.name == "日期":
        idx = "日期"
    else:
        idx = "date"
    df[dummy_col]=pd.to_datetime(df[dummy_col]).dt.to_period(_mapping[interval])
    return df.groupby([dummy_col])[cols].sum().rename_axis(idx)#.drop(columns=[dummy_col])


def irr(flow,init=None):
    def extract_cash_col(_cols):
        if _cols == china_bondflow_fields_s:
            return flow['本息合计']
        elif _cols == english_bondflow_fields_s: 
            return flow['cash']
        else:
            raise RuntimeError("Failed to match",_cols)

    cols = flow.columns.to_list()
    dates = flow.index.to_list()
    amounts = extract_cash_col(cols).to_list()
        
    if init is not None:
        invest_date,invest_amount = init
        dates = [invest_date]+dates
        amounts = [invest_amount]+amounts
    
    return xirr(np.array(dates), np.array(amounts))


def sum_fields_to_field(_df,cols,col):
    df = _df.copy()
    df[col] = df[cols].sum(axis=1)
    return df


def npv(_flow,**p):
    flow = _flow.copy()
    cols = flow.columns.to_list()
    idx_name = flow.index.name
    init_date,_init_amt = p['init']
    init_amt = _init_amt if _init_amt!=0.00 else 0.0001
    def _pv(_af):
        af = flow[_af]
        return xnpv(p['rate'],[init_date]+flow.index.to_list(),[-1*init_amt]+af.to_list())
    match (cols,idx_name):
        case (china_rental_flow,"日期"):
            return _pv("租金")
        case (english_rental_flow,"Date"):
            return _pv("Rental")
        case (english_mortgage_flow_fields,"Date"):
            sum_fields_to_field(flow,["Principal","Interest","Prepayment","Recovery"], "Cash")
            return _pv("Cash")
        case (china_bondflow_fields,"日期"):
            return _pv("本息合计")
        case (english_bondflow_fields,"Date"):
            return _pv("cash")
        case _:
            raise RuntimeError("Failed to match",cols,idx_name)


def update_deal(d,i,c):
    "A patch function to update a deal data component list in immuntable way"
    _d = d.copy()
    _d.pop(i)
    _d.insert(i,c)
    return _d

def mkDealsBy(d, m:dict)->dict:
    return { k:dataclasses.replace(d, **v) 
                for k,v in m.items()} 

class DC(Enum):  # TODO need to check with HS code
    DC_30E_360 = "DC_30E_360"
    DC_30Ep_360 ="DC_30Ep_360"
    DC_ACT_360  = "DC_ACT_360"
    DC_ACT_365A = "DC_ACT_365A"
    DC_ACT_365L = "DC_ACT_365L"
    DC_NL_365 = "DC_NL_365"
    DC_ACT_365F = "DC_ACT_365F"
    DC_ACT_ACT = "DC_ACT_ACT"
    DC_30_360_ISDA = "DC_30_360_ISDA"
    DC_30_360_German = "DC_30_360_German"
    DC_30_360_US  = "DC_30_360_US"

def str2date(x:str):
    return datetime.strptime(x, '%Y-%m-%d').date()

def normDate(x:str):
    if len(x)==8:
        return f"{x[:4]}-{x[4:6]}-{x[6:8]}"

def daysBetween(sd,ed):
    return (ed - sd).days

def guess_locale(x):
    accs = x['accounts']

    assert len(accs)>0,"Failed to identify via deal accounts result"

    acc_cols = set(list(accs.values())[0].columns.to_list())
    locale = None
    if acc_cols == set(["余额", "变动额", "备注"]):
        locale="cn"
    if acc_cols == set(["balance", "change", "memo"]):
        locale="en"
    return locale

def guess_pool_locale(x):
    if "cutoffDate" in x:
        return "english"
    elif '封包日' in x:
        return "chinese"
    else:
        raise RuntimeError("Failed to match {x} in guess pool locale")

def renameKs(m:dict,mapping,opt_key=False):
    for (o,n) in mapping:
        if opt_key and o not in m:
            continue
        m[n] = m[o]
        del m[o]
    return m

def subMap(m:dict,ks:list):
    ''' get a map subset by keys,if keys not found, supplied with default value '''
    return {k:m.get(k,defaultVal) for (k,defaultVal) in ks}

def subMap2(m:dict,ks:list):
    fieldNames = [ fName for (fName,fTargetName,fDefaultValue) in ks]
    _m = {k:m.get(k,defaultVal) for (k,_,defaultVal) in ks}
    _mapping = [ _[:2] for _ in ks ]
    return renameKs(_m,_mapping)

def mapValsBy(m:dict, f):
    assert isinstance(m, dict),"M is not a map"
    return {k: f(v) for k,v in m.items()}

def mapListValBy(m:dict, f):
    assert isinstance(m, dict),"M is not a map"
    return {k: [f(_v) for _v in v] for k,v in m.items()}

def applyFnToKey(m:dict, f, k, applyNone=False):
    assert isinstance(m, dict),f"{m} is not a map"
    assert k in m, f"{k} is not in map {m}"
    match (m[k],applyNone):
        case (None,True):
            m[k] = f(m[k])
        case (None,False):
            pass
        case (_, _):
            m[k] = f(m[k])
    return m

def renameKs2(m:dict,kmapping):
    assert isinstance(m, dict),"M is not a map"
    assert isinstance(kmapping, dict),f"Mapping is not a map: {kmapping}"
    assert set(m.keys()).issubset(set(kmapping.keys())),f"{m.keys()} not in {kmapping.keys()}"
    return {kmapping[k]:v for k,v in m.items()}

def ensure100(xs,msg=""):
    assert sum(xs)==1.0,f"Doesn't not sum up 100%: {msg}"

def guess_pool_flow_header(x,l):
    assert isinstance(x, dict), f"x is not a map but {x}, type:{type(x)}"
    match (x['tag'],l):
        case ('MortgageFlow','chinese'):
            return (china_mortgage_flow_fields_d,"日期")
        case ('MortgageFlow','english'):
            return (english_mortgage_flow_fields_d,"Date")
        case ('LoanFlow','chinese'):
            return (china_loan_flow_d,"日期")
        case ('LoanFlow','english'):
            return (english_loan_flow_d,"Date")
        case ('LeaseFlow','chinese'):
            return (china_rental_flow_d,"日期")
        case ('LeaseFlow','english'):
            return (english_rental_flow_d,"Date")
        case _:
            raise RuntimeError(f"Failed to match pool header with {x['tag']}{l}")

def uplift_m_list(l:list):
    return {k:v
            for m in l
            for k,v in m.items()}

def getValWithKs(m:dict,ks:list):
    ''' Get first available key/value in m'''
    if isinstance(m, dict):
        for k in ks:
            if k in m:
                return m[k]
    else:
        for k in ks:
            if hasattr(m, k):
                return getattr(m, k)
    return None

def _read_cf(x, lang):
    flow_header,idx = guess_pool_flow_header(x[0],lang)
    try:
        result = pd.DataFrame([_['contents'] for _ in x] , columns=flow_header)
    except ValueError as e:
        logging.error(f"Failed to match header:{flow_header} with {result[0]['contents']}")
    result = result.set_index(idx)
    result.index.rename(idx, inplace=True)
    result.sort_index(inplace=True)
    return result

def _read_asset_pricing(xs, lang):
    header = assetPricingHeader[lang]
    data = [ x['contents'] for x in xs]
    return pd.DataFrame(data, columns = header)

def mergeStrWithDict(s:str,m:dict) -> str:
    t = json.loads(s)
    t = t | m 
    return json.dumps(t)

def flow_by_scenario(rs, flowpath, node="col", rtn_df=True, ax = 1, rnd=2):
    "pull flows from multiple scenario"
    r = None
    if node=="col":
        r = {k:query(v,flowpath[:-1])[flowpath[-1]] for k,v  in rs.items()}    
    elif node=="idx":
        r = {k:query(v,flowpath[:-1]).loc[flowpath[-1]] for k,v  in rs.items()}    
    else:
        r = {k:query(v,flowpath) for k,v in rs.items()}
    if rtn_df:
        _vs = list(r.values())
        _ks = list(r.keys())
        r = pd.concat(_vs,keys=_ks,axis=ax) 
    return r

def positionFlow(x,m:dict,facePerPaper=100):
    _,_bflow = list(x['bonds'].items())[0]
    bflowHeader = _bflow.columns.to_list()

    def calcBondFlow(_bflow,factor):
        bflow = copy.deepcopy(_bflow)
        bflow[bflowHeader[0]] *= factor
        bflow[bflowHeader[1]] *= factor
        bflow[bflowHeader[2]] *= factor
        bflow[bflowHeader[4]] *= factor
        return bflow
        
    assert isinstance(m, dict),"Position info must be a map/dict"
    
    bOrignBal = {k:x['_deal']['contents']['bonds'][k]['bndOriginInfo']['originBalance'] for k,v in m.items()}
    bPapersPerBond = {k:v/facePerPaper for k,v in bOrignBal.items()}
    
    bflowFactor = {k:(v/bPapersPerBond[k]) for k,v in m.items()}
    
    return {bn:calcBondFlow(bf,bflowFactor[bn])  for bn,bf in x['bonds'].items() if bn in m}


from absbox.local.util import mkTag, DC, mkTs, guess_locale, readTagStr, subMap, subMap2, renameKs, ensure100, mapListValBy, uplift_m_list, mapValsBy, allList, getValWithKs, applyFnToKey
from absbox.local.base import *
from enum import Enum
import itertools
import functools
import logging

import pandas as pd
from pyspecter import query, S

def mkLiq(x):
    match x:
        case {"正常余额折价": cf, "违约余额折价": df}:
            return mkTag(("BalanceFactor", [cf, df]))
        case {"CurrentFactor": cf, "DefaultFactor": df}:
            return mkTag(("BalanceFactor", [cf, df]))
        case {"贴现计价": df, "违约余额回收率": r}:
            return mkTag(("PV", [df, r]))
        case {"PV": df, "DefaultRecovery": r}:
            return mkTag(("PV", [df, r]))
        case _:
            raise RuntimeError(f"Failed to match {x} in Liquidation Method")

def mkDatePattern(x):
    match x:
        case ["每月", _d]:
            return mkTag((datePattern["每月"], _d))
        case ["每年", _m, _d]:
            return mkTag((datePattern["每年"], [_m, _d]))
        case ["DayOfMonth", _d]:
            return mkTag(("DayOfMonth", _d))
        case ["MonthDayOfYear", _m, _d]:
            return mkTag(("MonthDayOfYear", _m, _d))
        case ["CustomDate", *_ds]:
            return mkTag(("CustomDate", _ds))
        case ["EveryNMonth", d, n]:
            return mkTag(("EveryNMonth", [d, n]))
        case ["AllDatePattern", *_dps]:
            return mkTag(("AllDatePattern", [ mkDatePattern(_) for _ in _dps]))
        case ["After", _d, dp] | ["之后", _d, dp]:
            return mkTag(("StartsExclusive", [ _d, mkDatePattern(dp) ]))
        case ["ExcludeDatePattern", _d, _dps] | ["排除", _d, _dps]:
            return mkTag(("Exclude", [ mkDatePattern(_d)
                                     , [mkDatePattern(_) for _ in _dps]]))
        case ["OffsetDateDattern", _dp, n] | ["平移", _dp, n]:
            return mkTag(("OffsetBy", [ mkDatePattern(_dp), n]))
        case _x if (_x in datePattern.values()):
            return mkTag((_x))
        case _x if (_x in datePattern.keys()):
            return mkTag((datePattern[x]))
        case _:
            raise RuntimeError(f"Failed to match {x}")

def getStartDate(x):
    match x:
        case {"封包日": a, "起息日": b, "首次兑付日": c, "法定到期日": d, "收款频率": pf, "付款频率": bf} | \
             {"cutoff": a, "closing": b, "firstPay": c, "stated": d, "poolFreq": pf, "payFreq": bf}:
            return (a,b)
        case {"归集日": (lastCollected, nextCollect), "兑付日": (pp, np), "法定到期日": c, "收款频率": pf, "付款频率": bf} | \
             {"collect": (lastCollected, nextCollect), "pay": (pp, np), "stated": c, "poolFreq": pf, "payFreq": bf}:
            return (lastCollected,pp)

def mkDate(x):
    match x:
        case {"封包日": a, "起息日": b, "首次兑付日": c, "法定到期日": d, "收款频率": pf, "付款频率": bf} | \
             {"cutoff": a, "closing": b, "firstPay": c, "stated": d, "poolFreq": pf, "payFreq": bf}:
            firstCollection = x.get("首次归集日", b)
            mr = x.get("循环结束日", None)
            return mkTag(("PreClosingDates", [a, b, mr, d, [firstCollection, mkDatePattern(pf)], [c, mkDatePattern(bf)]]))
        case {"归集日": (lastCollected, nextCollect), "兑付日": (pp, np), "法定到期日": c, "收款频率": pf, "付款频率": bf} | \
             {"collect": (lastCollected, nextCollect), "pay": (pp, np), "stated": c, "poolFreq": pf, "payFreq": bf}:
            mr = x.get("循环结束日", None)
            return mkTag(("CurrentDates", [[lastCollected, pp],
                                           mr,
                                           c,
                                           [nextCollect, mkDatePattern(pf)],
                                           [np, mkDatePattern(bf)]]))
        case {"回款日": cdays, "分配日": ddays, "封包日": cutoffDate, "起息日": closingDate} | \
                {"poolCollection": cdays, "distirbution": ddays, "cutoff": cutoffDate, "closing": closingDate}:
            return mkTag(("CustomDates", [cutoffDate, [mkTag(("PoolCollection", [cd, ""])) for cd in cdays], closingDate, [mkTag(("RunWaterfall", [dd, ""])) for dd in ddays]]))
        case _:
            raise RuntimeError(f"Failed to match:{x} in Dates")

def mkDsRate(x):
    if isinstance(x,float):
        return mkDs(("constant",x))
    else:
        return mkDs(x)

def mkFeeType(x):
    match x:
        case {"年化费率": [base, rate]} | {"annualPctFee": [base, rate]}:
            return mkTag(("AnnualRateFee", [mkDs(base) ,mkDsRate(rate)]))
        case {"百分比费率": [*desc, _rate]} | {"pctFee": [*desc, _rate]}:
            rate = mkDsRate(_rate)
            match desc:
                case ["资产池当期", "利息"] | ["poolCurrentCollection", "interest"] | ["资产池回款", "利息"]:
                    return mkTag(("PctFee", [mkTag(("PoolCurCollection", ["CollectedInterest"])), rate]))
                case ["已付利息合计", *bns] | ["paidInterest", *bns]:
                    return mkTag(("PctFee", [mkTag(("LastBondIntPaid", bns)), rate]))
                case ["已付本金合计", *bns] | ["paidPrincipal", *bns]:
                    return mkTag(("PctFee", [mkTag(("LastBondPrinPaid", bns)), rate]))
                case _:
                    raise RuntimeError(f"Failed to match on 百分比费率：{desc,rate}")
        case {"固定费用": amt} | {"fixFee": amt}:
            return mkTag(("FixFee", amt))
        case {"周期费用": [p, amt]} | {"recurFee": [p, amt]}:
            return mkTag(("RecurFee", [mkDatePattern(p), amt]))
        case {"自定义": fflow} | {"customFee": fflow}:
            return mkTag(("FeeFlow", mkTs("BalanceCurve", fflow)))
        case {"计数费用": [p, s, amt]} | {"numFee": [p, s, amt]}:
            return mkTag(("NumFee", [mkDatePattern(p), mkDs(s), amt]))
        case {"差额费用": [ds1, ds2]} | {"targetBalanceFee": [ds1, ds2]}:
            return mkTag(("TargetBalanceFee", [mkDs(ds1), mkDs(ds2)]))
        case _:
            raise RuntimeError(f"Failed to match on fee type:{x}")


def mkDateVector(x):
    match x:
        case dp if isinstance(dp, str):
            return mkTag(datePattern[dp])
        case [dp, *p] if (dp in datePattern.keys()):
            return mkTag((datePattern[dp], p))
        case _:
            raise RuntimeError(f"not match found: {x}")

def mkPoolSource(x):
    match x:
        case "利息" | "Interest" :
            return "CollectedInterest" 
        case "本金" | "Principal":
            return "CollectedPrincipal" 
        case "回收" | "Recovery" :
            return "CollectedRecoveries" 
        case "早偿" | "Prepayment" :
            return "CollectedPrepayment" 
        case "租金" | "Rental" :
            return "CollectedRental" 
        case _ :
            raise RuntimeError(f"not match found: {x} :make Pool Source")

def mkDs(x):
    "Making Deal Stats"
    match x:
        case ("债券余额",) | ("bondBalance",):
            return mkTag("CurrentBondBalance")
        case ("债券余额", *bnds) | ("bondBalance", *bnds):
            return mkTag(("CurrentBondBalanceOf", bnds))
        case ("初始债券余额",) | ("originalBondBalance",):
            return mkTag("OriginalBondBalance")
        case ("到期月份", bn) | ("monthsTillMaturity", bn):
            return mkTag(("MonthsTillMaturity", bn))
        case ("资产池余额",) | ("poolBalance",):
            return mkTag("CurrentPoolBalance")
        case ("初始资产池余额",) | ("originalPoolBalance",):
            return mkTag("OriginalPoolBalance")
        case ("资产池违约余额",) | ("currentPoolDefaultedBalance",):
            return mkTag("CurrentPoolDefaultedBalance")
        case ("资产池累计违约余额",) | ("cumPoolDefaultedBalance",):
            return mkTag("CumulativePoolDefaultedBalance")
        case ("资产池累计回收额",) | ("cumPoolRecoveries",):
            return mkTag("CumulativePoolRecoveriesBalance")
        case ("资产池累计违约率",) | ("cumPoolDefaultedRate",):
            return mkTag("CumulativePoolDefaultedRate")
        case ("资产池累计",*i) | ("cumPoolCollection",*i):
            return mkTag(("PoolCumCollection", [mkPoolSource(_) for _ in i] ))
        case ("资产池当期",*i) | ("curPoolCollection",*i):
            return mkTag(("PoolCurCollection", [mkPoolSource(_) for _ in i]))
        case ("债券系数",) | ("bondFactor",):
            return mkTag("BondFactor")
        case ("资产池系数",) | ("poolFactor",):
            return mkTag("PoolFactor")
        case ("债券利率",bn) | ("bondRate",bn):
            return mkTag(("BondRate", bn))
        case ("资产池利率",) | ("poolWaRate",):
            return mkTag("PoolWaRate")
        case ("所有账户余额",) | ("accountBalance"):
            return mkTag("AllAccBalance")
        case ("账户余额", *ans) | ("accountBalance", *ans):
            return mkTag(("AccBalance", ans))
        case ("账簿余额", *ans) | ("ledgerBalance", *ans):
            return mkTag(("LedgerBalance", ans))
        case ("债券待付利息", *bnds) | ("bondDueInt", *bnds):
            return mkTag(("CurrentDueBondInt", bnds))
        case ("债券已付利息", *bnds) | ("lastBondIntPaid", *bnds):
            return mkTag(("LastBondIntPaid", bnds))
        case ("债券低于目标余额", bn) | ("behindTargetBalance", bn):
            return mkTag(("BondBalanceGap", bn))
        case ("已提供流动性", *liqName) | ("liqCredit", *liqName):
            return mkTag(("LiqCredit", liqName))
        case ("债务人数量",) | ("borrowerNumber",):
            return mkTag(("CurrentPoolBorrowerNum"))
        case ("事件", loc, idx) | ("trigger", loc ,idx):
            dealCycleM = chinaDealCycle | englishDealCycle
            if not loc in dealCycleM:
                raise RuntimeError(f" {loc} not in map {dealCycleM}")
            return mkTag(("TriggersStatusAt",[dealCycleM[loc], idx]))
        case ("待付费用", *fns) | ("feeDue", *fns):
            return mkTag(("CurrentDueFee", fns))
        case ("已付费用", *fns) | ("lastFeePaid", *fns):
            return mkTag(("LastFeePaid", fns))
        case ("费用支付总额",cmt,*fns) | ("feeTxnAmount",cmt,*fns):
            return mkTag(("FeeTxnAmt",[fns, cmt]))
        case ("债券支付总额",cmt,*bns) | ("bondTxnAmount",cmt,*bns):
            return mkTag(("BondTxnAmt",[bns, cmt]))
        case ("账户变动总额",cmt,*ans) | ("accountTxnAmount",cmt,*ans):
            return mkTag(("AccTxnAmt",[ans, cmt]))
        case ("系数", ds, f) | ("factor", ds, f):
            return mkTag(("Factor", [mkDs(ds), f]))
        case ("Min", *ds):
            return mkTag(("Min", [mkDs(s) for s in ds]))
        case ("Max", *ds):
            return mkTag(("Max", [mkDs(s) for s in ds]))
        case ("合计", *ds) | ("sum", *ds):
            return mkTag(("Sum", [mkDs(_ds) for _ds in ds]))
        case ("差额", *ds) | ("substract", *ds):
            return mkTag(("Substract", [mkDs(_ds) for _ds in ds]))
        case ("常数", n) | ("constant", n):
            return mkTag(("Constant", n))
        case ("储备账户缺口", *accs) | ("reserveGap", *accs):
            return mkTag(("ReserveAccGap", accs))
        case ("自定义", n) | ("custom", n):
            return mkTag(("UseCustomData", n))
        case ("区间内",floor,cap,s) | ("floorCap",floor,cap,s):
            return mkTag(("FloorAndCap", [floor,cap,s]))
        case ("floorWith", ds1, ds2):
            return mkTag(("FloorWith", [mkDs(ds1),mkDs(ds2)]))
        case ("floorWithZero", ds1):
            return mkTag(("FloorWithZero", mkDs(ds1)))
        case ("capWith", ds1, ds2):
            return mkTag(("CapWith", [mkDs(ds1),mkDs(ds2)]))
        case legacy if (legacy in baseMap.keys()):
            return mkDs((legacy,))
        case _:
            raise RuntimeError(f"Failed to match DS/Formula: {x}")


def mkCurve(tag,xs):
    return mkTag((tag,xs))


def mkPre(p):
    def queryType(y):
        match y:
            case ("bondFactor",)|\
                 ("poolFactor",)|\
                 ("cumPoolDefaultedRate",)|\
                 ("资产池累计违约率",)|\
                 ("债券系数",)|\
                 ("资产池系数",):
                return "IfRate"
            case ("borrowerNumber",_)|\
                ("monthsTillMaturity",_):
                return "IfInt"
            case _:
                return "If"

    match p:
        case ["状态", _st] | ["status", _st]:
            return mkTag(("IfDealStatus", mkStatus(_st)))
        case ["同时满足", *_p] | ["all", *_p]:
            return mkTag(("All", [mkPre(p) for p in _p]))
        case ["任一满足", *_p] | ["any", *_p]:
            return mkTag(("Any", [mkPre(p) for p in _p]))
        case [ds, "=", 0]:
            return mkTag(("IfZero", mkDs(ds)))
        case [ds, b] | [ds, b] if isinstance(b, bool):
            return mkTag(("IfBool",[mkDs(ds), b]))
        case [ds1, op, ds2] if (isinstance(ds1, tuple) and isinstance(ds2, tuple)):
            q = queryType(ds1)
            return mkTag((f"{q}2", [op_map[op], mkDs(ds1), mkDs(ds2)]))
        case [ds, op, curve] if isinstance(curve, list):
            q = queryType(ds)
            return mkTag((f"{q}Curve", [op_map[op], mkDs(ds), mkCurve("ThresholdCurve",curve)]))
        case [ds, op, n]:
            q = queryType(ds)
            return mkTag((q, [op_map[op], mkDs(ds), n]))
        case [op, _d]:
            return mkTag(("IfDate",[op_map[op], _d]))
        case _:
            raise RuntimeError(f"Failed to match on Pre: {p}")


def mkAccInt(x):
    match x:
        case {"周期": _dp, "利率": idx, "利差": spd, "最近结息日": lsd} \
                | {"period": _dp,  "index": idx, "spread": spd, "lastSettleDate": lsd}:
            return mkTag(("InvestmentAccount", [idx, spd, lsd, mkDatePattern(_dp)]))
        case {"周期": _dp, "利率": br, "最近结息日": lsd} \
                | {"period": _dp, "rate": br, "lastSettleDate": lsd}:
            return mkTag(("BankAccount", [br, lsd,mkDatePattern(_dp)]))
        case None:
            return None
        case _:
            raise RuntimeError(
                f"Failed to match on account interest definition: {x}")


def mkAccType(x):
    match x:
        case {"固定储备金额": amt} | {"fixReserve": amt}:
            return mkTag(("FixReserve", amt))
        case {"目标储备金额": [base, rate]} | {"targetReserve": [base, rate]}:
            match base:
                case ["合计", *qs] | ["Sum", *qs]:
                    sumDs = [mkDs(q) for q in qs]
                    return mkTag(("PctReserve", [mkTag(("Sum", sumDs)), rate]))
                case _:
                    return mkTag(("PctReserve", [mkDs(base), rate]))
        case {"目标储备金额": {"公式": ds, "系数": rate}} | {"targetReserve": {"formula": ds, "factor": rate}}:
            return mkTag(("PctReserve", [mkDs(ds), rate]))
        case {"目标储备金额": {"公式": ds}} | {"targetReserve": {"formula": ds}}:
            return mkTag(("PctReserve", [mkDs(ds), 1.0]))
        case {"较高": [a, b]} | {"max": [a, b]}:
            return mkTag(("Max", [mkAccType(a), mkAccType(b)]))
        case {"较低": [a, b]} | {"min": [a, b]}:
            return mkTag(("Min", [mkAccType(a), mkAccType(b)]))
        case {"分段": [p, a, b]} | {"When": [p, a, b]}:
            return mkTag(("Either", [mkPre(p), mkAccType(a), mkAccType(b)]))
        case None:
            return None
        case _:
            raise RuntimeError(f"Failed to match {x} for account reserve type")


def mkAccTxn(xs):
    "AccTxn T.Day Balance Amount Comment"
    if xs is None:
        return None
    else:
        return [mkTag(("AccTxn", x)) for x in xs]


def mkAcc(an, x):
    match x:
        case {"余额": b, "类型": t, "计息": i, "记录": tx} | {"balance": b, "type": t, "interest": i, "txn": tx}:
            return {"accBalance": b, "accName": an, "accType": mkAccType(t), "accInterest": mkAccInt(i), "accStmt": mkAccTxn(tx)}

        case {"余额": b} | {"balance": b}:
            return mkAcc(an, x | {"计息": x.get("计息", None), "interest": x.get("interest", None), "记录": x.get("记录", None), "txn": x.get("txn", None), "类型": x.get("类型", None), "type": x.get("type", None)})
        case _:
            raise RuntimeError(f"Failed to match account: {an},{x}")


def mkBondType(x):
    match x:
        case {"固定摊还": schedule} | {"PAC": schedule}:
            return mkTag(("PAC", mkTag(("BalanceCurve", schedule))))
        case {"过手摊还": None} | {"Sequential": None}:
            return mkTag(("Sequential"))
        case {"锁定摊还": _after} | {"Lockout": _after}:
            return mkTag(("Lockout", _after))
        case {"权益": _} | {"Equity": _}:
            return mkTag(("Equity"))
        case _:
            raise RuntimeError(f"Failed to match bond type: {x}")


def mkBondRate(x):
    match x:
        case {"浮动": [_index, Spread, resetInterval], "日历": dc} | \
                {"floater": [_index, Spread, resetInterval], "dayCount": dc}:
            return mkTag(("Floater", [_index, Spread, mkDatePattern(resetInterval), dc, None, None]))
        case {"浮动": [_index, Spread, resetInterval]} | {"floater": [_index, Spread, resetInterval]}:
            return mkBondRate(x | {"日历": DC.DC_ACT_365F.value, "dayCount": DC.DC_ACT_365F.value})
        case {"固定": _rate, "日历": dc} | {"fix": _rate, "dayCount": dc}:
            return mkTag(("Fix", [_rate, dc]))
        case {"固定": _rate} | {"Fixed": _rate}:
            return mkTag(("Fix", [_rate, DC.DC_ACT_365F.value]))
        case {"调息": _rate, "幅度":spd, "调息日":dp} | {"StepUp": _rate, "Spread":spd, "When":dp}:
            return mkTag(("StepUpFix", [_rate, DC.DC_ACT_365F.value, mkDatePattern(dp), spd ]))
        case {"期间收益": _yield}:
            return mkTag(("InterestByYield", _yield))
        case _:
            raise RuntimeError(f"Failed to match bond rate type:{x}")


def mkBnd(bn, x):
    match x:
        case {"当前余额": bndBalance, "当前利率": bndRate, "初始余额": originBalance, "初始利率": originRate, "起息日": originDate, "利率": bndInterestInfo, "债券类型": bndType} | \
             {"balance": bndBalance, "rate": bndRate, "originBalance": originBalance, "originRate": originRate, "startDate": originDate, "rateType": bndInterestInfo, "bondType": bndType}:
            md = getValWithKs(x,["到期日","maturityDate"])
            lastAccrueDate = getValWithKs(x,["计提日","lastAccrueDate"])
            lastIntPayDate = getValWithKs(x,["付息日","lastIntPayDate"])
            return {"bndName": bn, "bndBalance": bndBalance, "bndRate": bndRate
                    , "bndOriginInfo": {"originBalance": originBalance, "originDate": originDate, "originRate": originRate} | {"maturityDate": md}
                    , "bndInterestInfo": mkBondRate(bndInterestInfo), "bndType": mkBondType(bndType)
                    , "bndDuePrin": 0, "bndDueInt": 0, "bndDueIntDate": lastAccrueDate
                    , "bndLastIntPayDate": lastIntPayDate}

        case _:
            raise RuntimeError(f"Failed to match bond:{bn},{x}:mkBnd")


def mkLiqMethod(x):
    match x:
        case ["正常|违约", a, b] | ["Current|Defaulted", a, b]:
            return mkTag(("BalanceFactor", [a, b]))
        case ["正常|拖欠|违约", a, b, c] | ["Cuurent|Delinquent|Defaulted", a, b, c]:
            return mkTag(("BalanceFactor2", [a, b, c]))
        case ["贴现|违约", a, b] | ["PV|Defaulted", a, b]:
            return mkTag(("PV", [a, b]))
        case ["贴现曲线", ts] | ["PVCurve", ts]:
            return mkTag(("PVCurve", mkTs("PricingCurve", ts)))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkLiqMethod")


def mkFeeCapType(x):
    match x:
        case {"应计费用百分比": pct} | {"duePct": pct}:
            return mkTag(("DuePct", pct))
        case {"应计费用上限": amt} | {"dueCapAmt": amt}:
            return mkTag(("DueCapAmt", amt))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkFeeCapType")


def mkPDA(x):
    match x:
        case {"公式": ds} | {"formula": ds}:
            return mkTag(("DS", mkDs(ds)))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkPDA")


def mkAccountCapType(x):
    match x:
        case {"余额百分比": pct} | {"balPct": pct}:
            return mkTag(("DuePct", pct))
        case {"金额上限": amt} | {"balCapAmt": amt}:
            return mkTag(("DueCapAmt", amt))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkAccountCapType")


def mkTransferLimit(x):
    match x:
        case {"余额百分比": pct} | {"balPct": pct}:
            return mkTag(("DuePct", pct))
        case {"金额上限": amt} | {"balCapAmt": amt}:
            return mkTag(("DueCapAmt", amt))
        case {"公式": "ABCD"}:
            return mkTag(("Formula", "ABCD"))
        case {"公式": formula} | {"formula": formula}:
            return mkTag(("DS", mkDs(formula)))
        case {"PDL贷记":an} | {"clearPDL":an}:
            return mkTag(("ClearPDL", an))
        case _:
            raise RuntimeError(f"Failed to match :{x}:mkTransferLimit")


def mkLiqRepayType(x):
    match x:
        case "余额" | "bal":
            return mkTag(("LiqBal"))
        case "费用" | "premium":
            return mkTag(("LiqPremium"))
        case "利息" | "int":
            return mkTag(("LiqInt"))
        case _:
            raise RuntimeError(f"Failed to match :{x}:Liquidation Repay Type")


def mkRateSwapType(rr,pr):
    def isFloater(y):
        if isinstance(y, tuple):
            return True
        return False
    match (isFloater(rr),isFloater(pr)):
        case (True,True):
            return mkTag(("FloatingToFloating"))
        case (False,True):
            return mkTag(("FixedToFloating"))
        case (True,False):
            return mkTag(("FloatingToFixed"))
        case _:
            raise RuntimeError(f"Failed to match :{rr,pr}:Interest Swap Type")

def mkRsBase(x):
    match x:
        case {"fixed":bal} | {"固定":bal}:
            return mkTag(("Fixed",bal))
        case {"formula": ds} | {"公式": ds}:
            return mkTag(("Base",mkDs(ds)))
        case _:
            raise RuntimeError(f"Failed to match :{x}:Interest Swap Base")


def mkRateSwap(x):
    match x:
        case {"settleDates":stl_dates,"payRate":p_rate,"receiveRate":r_rate 
             ,"base":base,"start":sd, "balance":bal,"lastSettleDate":lastStlDate,**p}:
            return mkTag(("RateSwap",
                                    {"rsType":mkRateSwapType(r_rate,p_rate),
                                    "settleDates":mkDatePattern(stl_dates),
                                    "notional":mkRsBase(base),
                                    "startDate":sd,
                                    "payingRate":p_rate,
                                    "receivingRate":r_rate,
                                    "refBalance":bal,
                                    "lastStlDate":lastStlDate,
                                    "netCash":p.get("netcash",0),
                                    "stmt":p.get("stmt",None)}))
        case _:
            raise RuntimeError(f"Failed to match :{x}:Interest Swap")

def mkBookLedgerType(x):
    match x:
        case ["PDL",defaults,ledgers]:
            return mkTag(("PDL",[mkDs(defaults)
                                 ,[[ln,mkDs(ds)] for ln,ds in ledgers]]))
        case _:
            raise RuntimeError(f"Failed to match :{x}:mkBookLedgerType")


def mkAction(x):
    match x:
        case ["账户转移", source, target] | ["transfer", source, target]:
            return mkTag(("Transfer", [source, target]))
        case ["按公式账户转移", _limit, source, target] | ["transferBy", _limit, source, target]:
            return mkTag(("TransferBy", [mkTransferLimit(_limit), source, target]))
        case ["簿记", bookLedgerType] | ["bookBy", bookLedgerType]:
            return mkTag(("BookBy", mkBookLedgerType(bookLedgerType)))
        case ["计提费用", *feeNames] | ["calcFee", *feeNames]:
            return mkTag(("CalcFee", feeNames))
        case ["计提利息", *bndNames] | ["calcInt", *bndNames]:
            return mkTag(("CalcBondInt", bndNames))
        case ["计提支付费用", source, target] | ["calcAndPayFee", source, target]:
            return mkTag(("CalcAndPayFee", [source, target]))
        case ["支付费用", source, target, _limit] | ["payFee", source, target, _limit]:
            limit = mkFeeCapType(_limit)
            return mkTag(("PayFeeBy", [limit, source, target]))        
        case ["支付费用", source, target] | ["payFee", source, target]:
            return mkTag(("PayFee", [source, target]))
        case ["支付费用限额", source, target, _limit] | ["payFeeBy", source, target, _limit]:
            limit = mkFeeCapType(_limit)
            return mkTag(("PayFeeBy", [limit, source, target]))
        case ["支付费用收益", source, target, _limit] | ["payFeeResidual", source, target, _limit]:
            limit = mkAccountCapType(_limit)
            return mkTag(("PayFeeResidual", [limit, source, target]))
        case ["支付费用收益", source, target] | ["payFeeResidual", source, target]:
            return mkTag(("PayFeeResidual", [None, source, target]))
        case ["计提支付利息", source, target] | ["accrueAndPayInt", source, target]:
            return mkTag(("AccrueAndPayInt", [source, target]))
        case ["支付利息", source, target] | ["payInt", source, target]:
            return mkTag(("PayInt", [source, target]))
        case ["支付本金", source, target, _limit] | ["payPrin", source, target, _limit]:
            limit = mkTransferLimit(_limit)
            return mkTag(("PayPrinBy", [limit, source, target]))
        case ["支付本金", source, target] | ["payPrin", source, target]:
            return mkTag(("PayPrin", [source, target]))
        case ["支付剩余本金", source, target] | ["payPrinResidual", source, target]:
            return mkTag(("PayPrinResidual", [source, target]))
        case ["支付期间收益", source, target]:
            return mkTag(("PayTillYield", [source, target]))
        case ["支付收益", source, target, limit] | ["payResidual", source, target, limit]:
            return mkTag(("PayResidual", [limit, source, target]))
        case ["支付收益", source, target] | ["payResidual", source, target]:
            return mkTag(("PayResidual", [None, source, target]))
        case ["储备账户转移", source, target, satisfy] | ["transferReserve", source, target, satisfy]:
            _map = {"源储备": "Source", "目标储备": "Target","Source":"Source","Target":"Target"}
            return mkTag(("TransferReserve", [_map[satisfy], source, target]))
        case ["出售资产", liq, target] | ["sellAsset", liq, target]:
            return mkTag(("LiquidatePool", [mkLiqMethod(liq), target]))
        case ["流动性支持", source, target, limit] | ["liqSupport", source, target, limit]:
            return mkTag(("LiqSupport", [   mkTransferLimit(limit), source, target]))
        case ["流动性支持", source, target] | ["liqSupport", source, target]:
            return mkTag(("LiqSupport", [None, source, target]))
        case ["流动性支持偿还", rpt, source, target] | ["liqRepay", rpt, source, target]:
            return mkTag(("LiqRepay", [None, mkLiqRepayType(rpt), source, target]))
        case ["流动性支持偿还", source, target] | ["liqRepay",  source, target]:
            return mkTag(("LiqRepay", [None, "bal", source, target]))
        case ["流动性支持报酬", source, target] | ["liqRepayResidual", source, target]:
            return mkTag(("LiqYield", [None, source, target]))
        case ["流动性支持计提", target] | ["liqAccrue", target]:
            return mkTag(("LiqAccrue", target))
        case ["条件执行", pre, *actions] | ["If", pre, *actions]:
            return mkTag(("ActionWithPre", [mkPre(pre), [mkAction(a) for a in actions] ] ))
        case ["条件执行2", pre, actions1, actions2] | ["IfElse", pre, actions1, actions2]:
            return mkTag(("ActionWithPre2", [mkPre(pre), [mkAction(a) for a in actions1], [mkAction(a) for a in actions2]] ))
        case ["购买资产", liq, source, _limit] | ["buyAsset", liq, source, _limit]:
            return mkTag(("BuyAsset", [_limit, mkLiqMethod(liq), source]))
        case ["更新事件", idx] | ["runTrigger", idx]:
            dealCycleM = chinaDealCycle | englishDealCycle
            return mkTag(("RunTrigger", ["InWF",idx]))
        case ["查看",*ds] | ["inspect",*ds]:
            return mkTag(("WatchVal",[None,[mkDs(_) for _ in ds]]))
        case _:
            raise RuntimeError(f"Failed to match :{x}:mkAction")


def mkStatus(x):
    match x:
        case "摊销" | "Amortizing":
            return mkTag(("Amortizing"))
        case "循环" | "Revolving":
            return mkTag(("Revolving"))
        case "加速清偿" | "Accelerated":
            return mkTag(("DealAccelerated", None))
        case "违约" | "Defaulted":
            return mkTag(("DealDefaulted", None))
        case "结束" | "Ended":
            return mkTag(("Ended"))
        case "设计" | "PreClosing":
            return mkTag(("PreClosing"))
        case _:
            raise RuntimeError(f"Failed to match :{x}:mkStatus")


def readStatus(x, locale):
    m = {"en": {'amort': "Amortizing", 'def': "Defaulted", 'acc': "Accelerated", 'end': "Ended",
                'pre': "PreClosing",'revol':"Revolving"}
        , "cn": {'amort': "摊销", 'def': "违约", 'acc': "加速清偿", 'end': "结束", 'pre': "设计","revol":"循环"}}
    match x:
        case {"tag": "Amortizing"}:
            return m[locale]['amort']
        case {"tag": "DealAccelerated"}:
            return m[locale]['acc']
        case {"tag": "DealDefaulted"}:
            return m[locale]['def']
        case {"tag": "Ended"}:
            return m[locale]['end']
        case {"tag": "PreClosing"}:
            return m[locale]['pre']
        case {"tag": "Revolving"}:
            return m[locale]['revol']
        case _:
            raise RuntimeError(
                f"Failed to read deal status:{x} with locale: {locale}")

def mkThreshold(x):
    match x:
        case ">":
            return "Above"
        case ">=":
            return "EqAbove"
        case "<":
            return "Below"
        case "<=":
            return "EqBelow"
        case _:
            raise RuntimeError(f"Failed to match :{x}:mkThreshold")


def _rateTypeDs(x):
    h = x[0]
    if h in set(["资产池累积违约率", "cumPoolDefaultedRate", "债券系数", "bondFactor", "资产池系数", "poolFactor"]):
        return True
    return False


def mkTrigger(x):
    match x:
        case {"condition":p,"effects":e,"status":st,"curable":c} | {"条件":p,"效果":e,"状态":st,"重置":c}:
            return {"trgCondition":mkPre(p)
                    ,"trgEffects":mkTriggerEffect(e)
                    ,"trgStatus":st
                    ,"trgCurable":c}
        case _:
            raise RuntimeError(f"Failed to match :{x}:mkTrigger")


def mkTriggerEffect(x):
    match x:
        case ("新状态", s) | ("newStatus", s):
            return mkTag(("DealStatusTo", mkStatus(s)))
        case ["计提费用", *fn] | ["accrueFees", *fn]:
            return mkTag(("DoAccrueFee", fn))
        case ["新增事件", trg] | ["newTrigger", trg]: # not implementd in Hastructure
            return mkTag(("AddTrigger", mkTrigger(trg)))
        case ["新储备目标",accName,newReserve] | ["newReserveBalance",accName,newReserve]:
            return mkTag(("ChangeReserveBalance",[accName, mkAccType(newReserve)]))
        case ["结果", *efs] | ["Effects", *efs]:
            return mkTag(("TriggerEffects", [mkTriggerEffect(e) for e in efs]))
        case _:
            raise RuntimeError(f"Failed to match :{x}:mkTriggerEffect")


def mkWaterfall(r, x):
    mapping = {
        "未违约": "Amortizing",
        "摊销": "Amortizing",
        "循环": "Revolving",
        "加速清偿": "DealAccelerated",
        "违约": "DealDefaulted",
        "未设立": "PreClosing",
    }
    if len(x) == 0:
        return {k: list(v) for k, v in r.items()}
    _k, _v = x.popitem()
    _w_tag = None
    match _k:
        case ("兑付日", "加速清偿") | ("amortizing", "accelerated"):
            _w_tag = f"DistributionDay (DealAccelerated Nothing)"
        case ("兑付日", "违约") | ("amortizing", "defaulted"):
            _w_tag = f"DistributionDay (DealDefaulted Nothing)"
        case ("兑付日", _st) | ("amortizing", _st):
            _w_tag = f"DistributionDay {mapping.get(_st,_st)}"
        case "兑付日" | "未违约" | "amortizing":
            _w_tag = f"DistributionDay Amortizing"
        case "清仓回购" | "cleanUp":
            _w_tag = "CleanUp"
        case "回款日" | "回款后" | "endOfCollection":
            _w_tag = f"EndOfPoolCollection"
        case "设立日" | "closingDay":
            _w_tag = f"OnClosingDay"
        case "默认" | "default":
            _w_tag = f"DefaultDistribution"
        case _:
            raise RuntimeError(f"Failed to match :{x}:mkWaterfall")
    r[_w_tag] = [mkAction(_a) for _a in _v]
    return mkWaterfall(r, x)

def mkRoundingType(x):
    match x:
        case ["floor",r]:
            return mkTag(("RoundFloor",r))
        case ["ceiling",r]:
            return mkTag(("RoundCeil",r))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkRoundingType")

def mkAssetRate(x):
    match x:
        case ["固定", r] | ["fix", r]:
            return mkTag(("Fix", r))
        case ["浮动", r, {"基准": idx, "利差": spd, "重置频率": p}]:
            return mkTag(("Floater", [idx, spd, r, freqMap[p], None]))
        case ["floater", r, {"index": idx, "spread": spd, "reset": p} as m]:
            _m = subMap(m,[("cap",None),("floor",None),("rounding",None)])
            _m = applyFnToKey(_m, mkRoundingType, 'rounding')
            return mkTag(("Floater2", [idx, spd, r, mkDatePattern(p), _m['floor'], _m['cap'],_m['rounding']]))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkAssetRate")

def mkAmortPlan(x) -> dict:
    match x:
        case "等额本息" | "Level" | "level":
            return mkTag("Level")
        case "等额本金" | "Even" | "even":
            return mkTag("Even")
        case "先息后本" | "I_P" | "i_p":
            return mkTag("I_P")
        case "等本等费" | "F_P" | "f_p":
            return mkTag("F_P")
        case _:
            raise RuntimeError(f"Failed to match AmortPlan {x}:mkAmortPlan")

def mkArm(x):
    match x:
        case {"initPeriod":ip}:
            fc = x.get("firstCap",None)
            pc = x.get("periodicCap",None)
            floor = x.get("lifeFloor",None)
            cap = x.get("lifeCap",None)
            return mkTag(("ARM",[ip,fc,pc,cap,floor]))
        case _:
            raise RuntimeError(f"Failed to match AmortPlan {x}:mkArm")

def mkAssetStatus(x):
    match x:
        case "正常"|"Current"|"current":
            return mkTag(("Current"))
        case "违约"|"Defaulted"|"defaulted":
            return mkTag(("Defaulted",None))
        case ("违约",d) |("Defaulted",d)|("defaulted",d):
            return mkTag(("Defaulted",d))
        case _:
            raise RuntimeError(f"Failed to match asset statuts {x}:mkAssetStatus")


def mkAsset(x):
    match x:
        case ["AdjustRateMortgage", {"originBalance": originBalance, "originRate": originRate, "originTerm": originTerm, "freq": freq, "type": _type, "originDate": startDate, "arm": arm}
             , {"currentBalance": currentBalance, "currentRate": currentRate, "remainTerm": remainTerms, "status": status}]:
            borrowerNum1 = x[2].get("borrowerNum", None)
            return mkTag(("AdjustRateMortgage",
                          [{"originBalance": originBalance,
                            "originRate": mkAssetRate(originRate),
                            "originTerm": originTerm,
                            "period": freqMap[freq],
                            "startDate": startDate,
                            "prinType": mkAmortPlan(_type)
                            } | mkTag("MortgageOriginalInfo"),
                            mkArm(arm),
                            currentBalance,
                            currentRate,
                            remainTerms,
                            borrowerNum1,
                            mkAssetStatus(status)]
            )) 
        case ["按揭贷款", {"放款金额": originBalance, "放款利率": originRate, "初始期限": originTerm, "频率": freq, "类型": _type, "放款日": startDate}, {"当前余额": currentBalance, "当前利率": currentRate, "剩余期限": remainTerms, "状态": status}] | \
                ["Mortgage", {"originBalance": originBalance, "originRate": originRate, "originTerm": originTerm, "freq": freq, "type": _type, "originDate": startDate}, {"currentBalance": currentBalance, "currentRate": currentRate, "remainTerm": remainTerms, "status": status}]:

            borrowerNum1 = x[2].get("borrowerNum", None)
            borrowerNum2 = x[2].get("借款数量", None)

            return mkTag(("Mortgage", [
                {"originBalance": originBalance,
                 "originRate": mkAssetRate(originRate),
                 "originTerm": originTerm,
                 "period": freqMap[freq],
                 "startDate": startDate,
                 "prinType": mkAmortPlan(_type)
                 } | mkTag("MortgageOriginalInfo"),
                currentBalance,
                currentRate,
                remainTerms,
                (borrowerNum1 or borrowerNum2),
                mkAssetStatus(status)]))
        case ["贷款", {"放款金额": originBalance, "放款利率": originRate, "初始期限": originTerm, "频率": freq, "类型": _type, "放款日": startDate}, {"当前余额": currentBalance, "当前利率": currentRate, "剩余期限": remainTerms, "状态": status}] \
                | ["Loan", {"originBalance": originBalance, "originRate": originRate, "originTerm": originTerm, "freq": freq, "type": _type, "originDate": startDate}, {"currentBalance": currentBalance, "currentRate": currentRate, "remainTerm": remainTerms, "status": status}]:
            return mkTag(("PersonalLoan", [
                {"originBalance": originBalance,
                 "originRate": mkAssetRate(originRate),
                 "originTerm": originTerm,
                 "period": freqMap[freq],
                 "startDate": startDate,
                 "prinType": mkAmortPlan(_type)
                 } | mkTag("LoanOriginalInfo"),
                currentBalance,
                currentRate,
                remainTerms,
                mkAssetStatus(status)]))
        case ["分期", {"放款金额": originBalance, "放款费率": originRate, "初始期限": originTerm, "频率": freq, "类型": _type, "放款日": startDate}, {"当前余额": currentBalance, "剩余期限": remainTerms, "状态": status}] \
                | ["Installment", {"originBalance": originBalance, "feeRate": originRate, "originTerm": originTerm, "freq": freq, "type": _type, "originDate": startDate}, {"currentBalance": currentBalance, "remainTerm": remainTerms, "status": status}]:
            return mkTag(("Installment", [
                {"originBalance": originBalance,
                 "originRate": mkAssetRate(originRate),
                 "originTerm": originTerm,
                 "period": freqMap[freq],
                 "startDate": startDate,
                 "prinType": mkAmortPlan(_type)
                 } | mkTag("LoanOriginalInfo"),
                currentBalance,
                remainTerms,
                mkAssetStatus(status)]))
        case ["租赁", {"固定租金": dailyRate, "初始期限": originTerm, "频率": dp, "起始日": startDate, "状态": status, "剩余期限": remainTerms}] \
                | ["Lease", {"fixRental": dailyRate, "originTerm": originTerm, "freq": dp, "originDate": startDate, "status": status, "remainTerm": remainTerms}]:
            return mkTag(("RegularLease", [{"originTerm": originTerm, "startDate": startDate, "paymentDates": mkDatePattern(dp), "originRental": dailyRate} | mkTag("LeaseInfo"), 0, remainTerms, _statusMapping[status]]))
        case ["租赁", {"初始租金": dailyRate, "初始期限": originTerm, "频率": dp, "起始日": startDate, "计提周期": accDp, "涨幅": rate, "状态": status, "剩余期限": remainTerms}] \
                | ["Lease", {"initRental": dailyRate, "originTerm": originTerm, "freq": dp, "originDate": startDate, "accrue": accDp, "pct": rate, "status": status, "remainTerm": remainTerms}]:

            dailyRatePlan = None
            _stepUpType = "curve" if isinstance(rate, list) else "constant"
            if _stepUpType == "constant":
                dailyRatePlan = mkTag(
                    ("FlatRate", [mkDatePattern(accDp), rate]))
            else:
                dailyRatePlan = mkTag(
                    ("ByRateCurve", [mkDatePattern(accDp), rate]))
            return mkTag(("StepUpLease", [{"originTerm": originTerm, "startDate": startDate, "paymentDates": mkDatePattern(dp), "originRental": dailyRate} | mkTag("LeaseInfo"), dailyRatePlan, 0, remainTerms, mkAssetStatus(status)]))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkAsset")


def identify_deal_type(x):
    match x:
        case {"pool": {"assets": [{'tag': 'PersonalLoan'}, *rest]}}:
            return "LDeal"
        case {"pool": {"assets": [{'tag': 'Mortgage'}, *rest]}}:
            return "MDeal"
        case {"pool": {"assets": [{'tag': 'AdjustRateMortgage'}, *rest]}}:
            return "MDeal"
        case {"pool": {"assets": [], "futureCf": cfs}} if cfs[0]['tag'] == 'MortgageFlow':
            return "MDeal"
        case {"pool": {"assets": [{'tag': 'Installment'}, *rest]}}:
            return "IDeal"
        case {"pool": {"assets": [{'tag': 'Lease'}, *rest]}} | {"pool": {"assets": [{'tag': 'RegularLease'}, *rest]}}:
            return "RDeal"
        case {"pool": {"assets": [{'tag': 'StepUpLease'}, *rest]}}:
            return "RDeal"
        case _:
            raise RuntimeError(f"Failed to identify deal type {x}")


def mkCallOptions(x):
    match x:
        case {"资产池余额": bal} | {"poolBalance": bal}:
            return mkTag(("PoolBalance", bal))
        case {"债券余额": bal} | {"bondBalance": bal}:
            return mkTag(("BondBalance", bal))
        case {"资产池余额剩余比率": factor} | {"poolFactor": factor}:
            return mkTag(("PoolFactor", factor))
        case {"债券余额剩余比率": factor} | {"bondFactor": factor}:
            return mkTag(("BondFactor", factor))
        case {"指定日之后": d} | {"afterDate": d}:
            return mkTag(("AfterDate", d))
        case {"任意满足": xs} | {"or": xs}:
            return mkTag(("Or", [mkCallOptions(_x) for _x in xs]))
        case {"全部满足": xs} | {"and": xs}:
            return mkTag(("And", [mkCallOptions(_x) for _x in xs]))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkCallOptions")


def mkAssumption(x) -> dict:
    assert isinstance(x, dict),f"An assumption should be a map/dict,but got {x}, type:{type(x)}"
    match x:
        case {"CPR": cpr} if isinstance(cpr, list):
            return mkTag(("PrepaymentVec", cpr))
        case {"CDR": cdr} if isinstance(cdr, list):
            return mkTag(("DefaultVec", cdr))
        case {"CPR": cpr}:
            return mkTag(("PrepaymentCPR", cpr))
        case {"CPR调整": [*cprAdj, ed]} | {"CPRAdjust": [*cprAdj, ed]}:
            return mkTag(("PrepaymentFactors", mkTs("FactorCurveClosed", [cprAdj, ed])))
        case {"CDR": cdr}:
            return mkTag(("DefaultCDR", cdr))
        case {"CDR调整": [*cdrAdj, ed]} | {"CDRAdjust": [*cdrAdj, ed]}:
            return mkTag(("DefaultFactors", mkTs("FactorCurveClosed", [cdrAdj, ed])))
        case {"DefaultedRecovery":[r,lag,timing]} | {"已违约回收":[r,lag,timing]}:
            ensure100(timing,"Sum of Recovery Timing")
            return mkTag(("DefaultedRecovery",[r,lag,timing]))
        case {"回收": (rr, rlag)} | {"Recovery": (rr, rlag)}:
            return mkTag(("Recovery", (rr, rlag)))
        case {"利率": [idx, rate]} if isinstance(rate, float):
            return mkTag(("InterestRateConstant", [idx, rate]))
        case {"Rate": [idx, rate]} if isinstance(rate, float):
            return mkTag(("InterestRateConstant", [idx, rate]))
        case {"利率": [idx, *rateCurve]} | {"Rate": [idx, *rateCurve]}:
            curve = mkTag(("IRateCurve", [ [t,v] for (t,v) in rateCurve]))
            return mkTag(("InterestRateCurve", [idx, curve]))
        case {"清仓": opts} | {"CleanUp": opts}:
            return mkTag(("CallWhen", [mkCallOptions(co) for co in opts]))
        case {"停止": d} | {"StopAt": d}:
            return mkTag(("StopRunBy", d))
        case {"租赁截止日": d} | {"LeaseProjectEnd": d}:
            return mkTag(("LeaseProjectionEnd", d))
        case {"租赁年涨幅": r} | {"LeaseAnnualIncreaseRate": r} if not isinstance(r, list):
            return mkTag(("LeaseBaseAnnualRate", r))
        case {"租赁年涨幅": r} | {"LeaseAnnualIncreaseRate": r}:
            return mkTag(("LeaseBaseCurve", mkTs("FloatCurve", r)))
        case {"租赁间隔": n} | {"LeaseGapDays": n}:
            return mkTag(("LeaseGapDays", n))
        case {"租赁间隔表": (tbl, n)} | {"LeaseGapDaysByAmount": (tbl, n)}:
            return mkTag(("LeaseGapDaysByAmount", [tbl, n]))
        case {"查看":inspects} | {"Inspect":inspects}:
            inspectVars = [ [mkDatePattern(dp),mkDs(ds)] for dp,ds in inspects ]
            return mkTag(("InspectOn", inspectVars))
        case {"FinancialReports": {"dates":dp} } | {"财务报表": {"日期":dp}} : 
            return mkTag(("BuildFinancialReport", mkDatePattern(dp)))
        case {"RevolvingAssets": [rpool,rassumps]} | {"循环资产": [rpool,rassumps]}:
            assumps = [ mkAssumption(ra) for ra in rassumps ]
            return mkTag(("AvailableAssets",[mkRevolvingPool(rpool), assumps]))
        case {"预计费用":[fname,curve]} | {"EstimateExpense":[fname,curve]}:
            return mkTag(("ProjectedExpense",[fname, mkTs("BalanceCurve",curve)]))
        case _:
            raise RuntimeError(f"Failed to match {x}:Assumption")

def mkAssetUnion(x):
    match x[0]:
        case "AdjustRateMortgage" | "Mortgage" | "按揭贷款" :
            return mkTag(("MO",mkAsset(x)))
        case "贷款" | "Loan" : 
            return mkTag(("LO",mkAsset(x)))
        case "分期" | "Installment" : 
            return mkTag(("IL",mkAsset(x)))
        case "租赁" | "Lease" : 
            return mkTag(("LS",mkAsset(x)))
        case _:
            raise RuntimeError(f"Failed to match AssetUnion {x}")

def mkRevolvingPool(x):
    match x:
        case ["constant",asts]|["固定",asts]:
            return mkTag(("ConstantAsset",[ mkAssetUnion(_) for _ in asts]))
        case ["static",asts]|["静态",asts]:
            return mkTag(("StaticAsset",[ mkAssetUnion(_) for _ in asts]))
        case ["curve",astsWithDates]|["曲线",astsWithDates]:
            assetCurve = [ [d, [mkAssetUnion(a) for a in asts]] for (d,asts) in astsWithDates ]            
            return mkTag(("AssetCurve",assetCurve))

def mkAssumpList(xs):
    assert isinstance(xs, list), f"Assumption should be a list, but got {xs}"
    return [ mkAssumption(x) for x in xs ]

def mkAssumption2(x) -> dict:
    match x:
        case ["ByIndex", assetAssumpList, dealAssump] | ["明细", assetAssumpList, dealAssump]:
            return mkTag(("ByIndex"
                         , [[(ids, mkAssumpList(aps)) for ids, aps in assetAssumpList]
                         , mkAssumpList(dealAssump)]))
        case xs if isinstance(xs, list):
            return mkTag(("PoolLevel", mkAssumpList(xs)))
        case None:
            return mkTag(("PoolLevel", []))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkAssumption2, type:{type(x)}")


def mkPool(x):
    mapping = {"LDeal": "LPool", "MDeal": "MPool",
               "IDeal": "IPool", "RDeal": "RPool"}
    match x:
        case {"清单": assets, "封包日": d} | {"assets": assets, "cutoffDate": d}:
            _pool = {"assets": [mkAsset(a) for a in assets], "asOfDate": d}
            _pool_asset_type = identify_deal_type({"pool": _pool})
            return mkTag((mapping[_pool_asset_type], _pool))
        case _:
            raise RuntimeError(f"Failed to match {x}:mkPool")


def mkCustom(x):
    match x:
        case {"常量": n} | {"Constant": n}:
            return mkTag(("CustomConstant", n))
        case {"余额曲线": ts} | {"BalanceCurve": ts}:
            return mkTag(("CustomCurve", mkTs("BalanceCurve", ts)))
        case {"公式": ds} | {"Formula": ds}:
            return mkTag(("CustomDS", mkDs(ds)))


def mkLiqProviderType(x):
    match x:
        case {"总额度": amt} | {"total": amt}:
            return mkTag(("FixSupport"))
        case {"日期": dp, "限额": amt} | {"reset": dp, "quota": amt}:
            return mkTag(("ReplenishSupport", [mkDatePattern(dp), amt]))
        case {"公式": ds, "系数":pct} | {"formula":ds, "pct":pct}:
            return mkTag(("ByPct", [mkDs(ds), pct]))
        case {}:
            return mkTag(("UnLimit"))
        case _:
            raise RuntimeError(f"Failed to match LiqProvider Type：{x}")
        
def mkLiqProviderRate(x):
    match x:
        case {"fixRate":r ,"rateAccDates":rateAccDates,"lastAccDate":lastAccDate} | \
              {"固定利率":r ,"结息日":rateAccDates,"上次结息日":lastAccDate} :
            return mkTag(("FixRate",[mkDatePattern(rateAccDates),r,lastAccDate]))
        case _:
            return None

def buildOptionalLiqProvider(m):
    optionalFields = ["liqDueInt","liqDuePremium","liqRate","liqPremium","liqStmt"]
    r = {}
    for k in optionalFields:
        if k in m:
            if k in set(["liqRate","liqPremium"]):
                r[k] = mkLiqProviderRate(m[k])
            else:
                r[k] = m[k]
        else:
            r[k] = None
    return r

def mkLiqProvider(n, x):
    opt_fields = {"liqCredit":0,"liqDueInt":None,"liqDuePremium":None
                 ,"liqRate":None,"liqPremiumRate":None,"liqStmt":None}

    x_transformed = renameKs(x,[("已提供","liqCredit"),("应付利息","liqDueInt"),("应付费用","liqDuePremium")
                                ,("利率","liqRate"),("费率","liqPremium"),("记录","liqStmt")]
                                ,opt_key=True)
    r = None
    match x_transformed :
        case {"类型": "无限制", "起始日": _sd, **p} | {"type": "Unlimited", "start": _sd, **p}:
            r = {"liqName": n, "liqType": mkLiqProviderType({})
                    ,"liqBalance": None, "liqStart": _sd
                    ,"liqRate":mkLiqProviderRate(p)} | buildOptionalLiqProvider(p)
        case {"类型": _sp, "额度": _ab, "起始日": _sd, **p} \
                | {"type": _sp, "lineOfCredit": _ab, "start": _sd, **p}:
            r = {"liqName": n, "liqType": mkLiqProviderType(_sp)
                    ,"liqBalance": _ab,  "liqStart": _sd
                    ,"liqRate":mkLiqProviderRate(p)} | buildOptionalLiqProvider(p)
        case {"额度": _ab, "起始日": _sd, **p} | {"lineOfCredit": _ab, "start": _sd, **p}:
            r = {"liqName": n, "liqType": mkTag(("FixSupport"))
                    ,"liqBalance": _ab,  "liqStart": _sd
                    ,"liqRate":mkLiqProviderRate(p)} | buildOptionalLiqProvider(p)
        case _:
            raise RuntimeError(f"Failed to match LiqProvider:{x}")

    if r is not None:
       return opt_fields | r 

def mkLedger(n, x):
    match x:
        case {"balance":bal,"txn":_tx} | {"余额":bal,"记录":_tx}:
            tx = mkAccTxn(_tx)
            return {"ledgName":n,"ledgBalance":bal,"ledgStmt":tx}
        case _:
            raise RuntimeError(f"Failed to match Ledger:{x}")

def mkCf(x):
    if len(x) == 0:
        return None
    else:
        return [mkTag(("MortgageFlow", _x+[0.0]*5+[None])) for _x in x]


def mkCollection(x):
    match x :
        case [s,acc] if isinstance(acc, str):
            return mkTag(("Collect",[poolSourceMapping[s],acc]))
        case [s,pcts] if isinstance(pcts, list):
            return mkTag(("CollectByPct" ,[poolSourceMapping[s] ,pcts]))
        case _:
            raise RuntimeError(f"Failed to match collection rule {x}")

def mk(x):
    match x:
        case ["资产", assets]:
            return {"assets": [mkAsset(a) for a in assets]}
        case ["账户", accName, attrs] | ["account", accName, attrs]:
            return {accName: mkAcc(accName, attrs)}

def mkFee(x,fsDate=None):
    match x :
        case {"name":fn, "type": feeType, **fi}:
            opt_fields = subMap(fi, [("feeStart",fsDate),("feeDueDate",None),("feeDue",0),
                                    ("feeArrears",0),("feeLastPaidDate",None)])
            return  {"feeName": fn, "feeType": mkFeeType(feeType)} | opt_fields
        case {"名称":fn , "类型": feeType, **fi}:
            opt_fields = subMap2(fi, [("起算日","feeStart",fsDate),("计算日","feeDueDate",None),("应计费用","feeDue",0),
                                      ("拖欠","feeArrears",0),("上次缴付日期","feeLastPaidDay",None)])
            return  {"feeName": fn, "feeType": mkFeeType(feeType)} | opt_fields
        case _:
            raise RuntimeError(f"Failed to match fee: {x}")

def mkPricingAssump(x):
    match x:
        case {"贴现日": pricingDay, "贴现曲线": xs} | {"PVDate": pricingDay, "PVCurve": xs}:
            return mkTag(("DiscountCurve", [pricingDay, mkTs("IRateCurve", xs)]))
        case {"债券": bnd_with_price, "利率曲线": rdps} | {"bonds": bnd_with_price, "curve": rdps}:
            return mkTag(("RunZSpread", [mkTs("IRateCurve", rdps), bnd_with_price]))
        case _:
            raise RuntimeError(f"Failed to match pricing assumption: {x}")

def readPricingResult(x, locale) -> dict:
    if x is None:
        return None
    h = None

    tag = list(x.values())[0]["tag"]
    if tag == "PriceResult":
        h = {"cn": ["估值", "票面估值", "WAL", "久期", "凸性", "应计利息"],
             "en": ["pricing", "face", "WAL", "duration", "convexity", "accure interest"]}
    elif tag == "ZSpread":
        h = {"cn": ["静态利差"], "en": ["Z-spread"]}
    else:
        raise RuntimeError(
            f"Failed to read princing result: {x} with tag={tag}")

    return pd.DataFrame.from_dict({k: v['contents'] for k, v in x.items()}, orient='index', columns=h[locale]).sort_index()


def readRunSummary(x, locale) -> dict:
    def filter_by_tags(xs, tags):
        tags_set = set(tags)
        return [ x for x in xs if x['tag'] in tags_set]

    r = {}
    if x is None:
        return None

    bndStatus = {'cn': ["本金违约", "利息违约", "起算余额"]
                ,'en': ["Balance Defaults", "Interest Defaults", "Original Balance"]}
    bond_defaults = [(_['contents'][0], _['tag'], _['contents'][1], _['contents'][2])
                     for _ in x if _['tag'] in set(['BondOutstanding', 'BondOutstandingInt'])]
    _fmap = {"cn": {'BondOutstanding': "本金违约", "BondOutstandingInt": "利息违约"}
            ,"en": {'BondOutstanding': "Balance Defaults", "BondOutstandingInt": "Interest Defaults"}}
    bndNames = set([y[0] for y in bond_defaults])
    bndSummary = pd.DataFrame(columns=bndStatus[locale], index=list(bndNames))
    for bn, amt_type, amt, begBal in bond_defaults:
        bndSummary.loc[bn][_fmap[locale][amt_type]] = amt
        bndSummary.loc[bn][bndStatus[locale][2]] = begBal
    bndSummary.fillna(0, inplace=True)
    bndSummary["Total"] = bndSummary[bndStatus[locale][0]] + \
        bndSummary[bndStatus[locale][1]]

    r['bonds'] = bndSummary

    dealStatusLog = {'cn': ["日期", "旧状态", "新状态"], 'en': ["Date", "From", "To"]}
    status_change_logs = [(_['contents'][0], readStatus(_['contents'][1], locale), readStatus(_['contents'][2], locale))
                          for _ in x if _['tag'] in set(['DealStatusChangeTo'])]
    r['status'] = pd.DataFrame(data=status_change_logs, columns=dealStatusLog[locale])

    # inspection variables
    def uplift_ds(df):
        ds_name = readTagStr(df['DealStats'].iloc[0])
        df.drop(columns=["DealStats"],inplace=True)
        df.rename(columns={"Value":ds_name},inplace=True)
        df.set_index("Date",inplace=True)
        return df
    inspect_vars = filter_by_tags(x, ["InspectBal","InspectBool"])
    inspect_df = pd.DataFrame(data = [ (c['contents'][0],str(c['contents'][1]),c['contents'][2]) for c in inspect_vars ]
                              ,columns = ["Date","DealStats","Value"])
    grped_inspect_df = inspect_df.groupby("DealStats")

    r['inspect'] = {readTagStr(k):uplift_ds(v) for k,v in grped_inspect_df}
    
    # build financial reports
    def mapItem(z):
        match z:
            case {"tag":"Item","contents":[accName,accBal]}:
                return {accName:accBal}
            case {"tag":"ParentItem","contents":[accName,subItems]}:
                items = [ mapItem(i) for i in subItems]
                return {accName : items}

    def buildBalanceSheet(bsData):
        bsRptDate = bsData.pop("reportDate")
        bs = mapListValBy(bsData, mapItem)
        return mapValsBy(bs, uplift_m_list) | {"reportDate":bsRptDate}

    def buildBsType(yname, y:dict)-> pd.DataFrame:
        mi = pd.MultiIndex.from_product([[yname],y.keys()])
        d = y.values()
        return pd.DataFrame(d, index=mi).T
    
    def buildBS(bs):
        bs_df = pd.concat([  buildBsType(k,v)  for k,v in bs.items() if k!="reportDate"],axis=1)
        bs_df['reportDate'] = bs['reportDate']
        return bs_df.set_index("reportDate")

    def buildCashReport(cashData):
        sd = cashData.pop('startDate')
        ed = cashData.pop('endDate')
        net = cashData.pop('net')
        cashList = mapListValBy(cashData, mapItem)
        cashMap = {k:uplift_m_list(v) for k,v in cashList.items() }
        cashMap = pd.concat([  buildBsType(k,v) for k,v in cashMap.items() ],axis=1)
        cashMap['startDate'] = sd
        cashMap['endDate'] = ed
        cashMap['Net'] = net
        return cashMap.set_index(["startDate","endDate"])

    balanceSheetIdx = 2
    cashReportIdx = 3
    rpts = [ _['contents'] for  _ in  (filter_by_tags(x, ["FinancialReport"])) ]
    if rpts:
        r['report'] = {}
        r['report']['balanceSheet'] = pd.concat([buildBS(buildBalanceSheet(rpt[balanceSheetIdx])) for rpt in rpts])
        r['report']['cash'] = pd.concat([buildCashReport(rpt[cashReportIdx]) for rpt in rpts])[["inflow","outflow","Net"]]
    

    return r


def aggAccs(x, locale):
    _header = {
        "cn": {"idx": "日期", "change": "变动额", "bal": ("期初余额", '余额', "期末余额")}
        ,"en": {"idx": "date", "change": "change", "bal": ("begin balance", 'balance', "end balance")}
    }

    header = _header[locale]
    agg_acc = {}
    for k, v in x.items():
        acc_by_date = v.groupby(header["idx"])
        acc_txn_amt = acc_by_date.agg(change=(header["change"], sum)).rename(columns={"change":header["change"]})
        
        ending_bal_column = acc_by_date.last()[header["bal"][1]].rename(header["bal"][2])
        begin_bal_column = ending_bal_column.shift(1).rename(header["bal"][0])
        
        agg_acc[k] = acc_txn_amt.join([begin_bal_column, ending_bal_column])
        if agg_acc[k].empty:
            agg_acc[k].columns = header["bal"][0], header['change'], header["bal"][2]
            continue
        fst_idx = agg_acc[k].index[0]
        agg_acc[k].at[fst_idx, header["bal"][0]] = round(agg_acc[k].at[fst_idx,  header["bal"][2]] - agg_acc[k].at[fst_idx, header['change']], 2)
        agg_acc[k] = agg_acc[k][[header["bal"][0], header['change'], header["bal"][2]]]

    return agg_acc


def readIssuance(pool):
    _map = {'cn': "发行", 'en': "Issuance"}

    lang_flag = None
    if '发行' in pool.keys():
        lang_flag = 'cn'
    elif 'Issuance' in pool.keys():
        lang_flag = 'en'
    else:
        return None

    validIssuanceFields = {
        "资产池规模": "IssuanceBalance",
        "IssuanceBalance": "IssuanceBalance"
    }

    r = {}
    for k, v in pool[_map[lang_flag]].items():
        if k in validIssuanceFields:
            r[validIssuanceFields[k]] = v
        else:
            logging.warning(
                "Key {k} is not in pool fields {validIssuanceFields.keys()}")
    return r


def show(r, x="full"):
    ''' show cashflow of SPV during the projection '''
    def _map(y):
        if y == 'cn':
            return {"agg_accounts": "账户", "fees": "费用", "bonds": "债券", "pool": "资产池", "idx": "日期"}
        else:
            return {"agg_accounts": "Accounts", "fees": "Fees", "bonds": "Bonds", "pool": "Pool", "idx": "date"}

    _comps = ['agg_accounts', 'fees', 'bonds']

    dfs = {c: pd.concat(r[c].values(), axis=1, keys=r[c].keys())
           for c in _comps if r[c]}

    locale = guess_locale(r)
    _m = _map(locale)

    dfs2 = {}

    for k, v in dfs.items():
        dfs2[_m[k]] = pd.concat([v], keys=[_m[k]], axis=1)

    agg_pool = pd.concat([r['pool']['flow']], axis=1, keys=[_m["pool"]])
    agg_pool = pd.concat([agg_pool], axis=1, keys=[_m["pool"]])

    _full = functools.reduce(lambda acc, x: acc.merge(
        x, how='outer', on=[_m["idx"]]), [agg_pool]+list(dfs2.values()))

    match x:
        case "full":
            return _full.loc[:, [_m["pool"]]+list(dfs2.keys())].sort_index()
        case "cash":
            return None  # ""





import sys,warnings

if (sys.version_info.major >= 3 and sys.version_info.minor < 10):
    raise ImportError("AbsBox support Python with version 3.10+ only")

from absbox.client import API
from absbox.local.util import guess_pool_flow_header
from absbox.local.base import *
from absbox.local.china import 信贷ABS,SPV
from absbox.local.cmp import comp_engines
from absbox.local.generic import Generic


from importlib.metadata import version

__version__ = version("absbox")